from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional

import jax
import jax.numpy as jnp
from pantea.logger import logger
from pantea.pytree import BaseJaxPytreeDataClass, register_jax_pytree_node
from pantea.types import Array, Dtype, default_dtype


@jax.jit
def _apply_pbc(dx: Array, lattice: Array) -> Array:
    """Apply periodic boundary condition (PBC) along x,y, and z directions."""
    box = lattice.diagonal()
    dx = jnp.where(dx > 0.5 * box, dx - box, dx)
    dx = jnp.where(dx < -0.5 * box, dx + box, dx)
    return dx


@jax.jit
def _shift_inside_box(positions: Array, lattice: Array) -> Array:
    """Shift the input atom position inside the PBC simulation box."""
    box = lattice.diagonal()
    return jnp.remainder(positions, box)


@dataclass
class Box(BaseJaxPytreeDataClass):
    """
    Simulation box is used to implement periodic boundary
    conditions (PBC) in the presence of a lattice matrix.

    .. warning::
        The current implementation only works for orthogonal cells
        and does not support triclinic cells.
    """

    lattice: Array

    def __post_init__(self) -> None:
        """Post initialize simulation box (super-cell)."""
        logger.debug(f"Initializing {self}")
        self._assert_jit_static_attributes()
        self._assert_jit_dynamic_attributes(expected=("lattice",))

    @classmethod
    def from_list(
        cls,
        data: List[float],
        dtype: Optional[Dtype] = None,
    ) -> Box:
        if dtype is None:
            dtype = default_dtype.FLOATX
        lattice = jnp.array(data, dtype=dtype).reshape(3, 3)
        return Box(lattice)

    @jax.jit
    def apply_pbc(self, dx: Array) -> Array:
        """
        Apply the periodic boundary condition (PBC) on the provided position array.

        For this method to function correctly, it is essential that all atoms are
        initially positioned within the boundaries of the box.
        Otherwise, the results may not be as anticipated.

        :param dx: positional difference
        :type dx: Array
        :return: PBC applied input
        :rtype: Optional[Array]
        """
        return _apply_pbc(dx, self.lattice)

    @jax.jit
    def shift_inside_box(self, positions: Array) -> Array:
        """
        Adjust the coordinates of the input atoms to ensure they fall
        within the boundaries of the simulation box that implements
        periodic boundary conditions (PBC).

        :param x: atom positions
        :type x: Array
        :return: shifted atom positions
        :rtype: Array
        """
        logger.debug("Shift all atoms inside the simulation box")
        return _shift_inside_box(positions, self.lattice)

    @property
    def lx(self) -> Array:
        """Return length of cell in x-direction."""
        return self.lattice[0, 0]

    @property
    def ly(self) -> Array:
        """Return length of cell in y-direction."""
        return self.lattice[1, 1]

    @property
    def lz(self) -> Array:
        """Return length of cell in z-direction."""
        return self.lattice[2, 2]

    @property
    def length(self) -> Array:
        """Return length of cell in x, y, and z-directions."""
        return self.lattice.diagonal()

    @property
    def volume(self) -> Array:
        """Return volume of the box."""
        return jnp.prod(self.length)

    @property
    def dtype(self) -> Dtype:
        return self.lattice.dtype

    def __hash__(self) -> int:
        """Enforce to use the parent class's hash method (JIT)."""
        return super().__hash__()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}"
            f"(lattice={self.lattice}, dtype={self.dtype})"
        )


register_jax_pytree_node(Box)



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

_optional_attr = field(default=None, repr=False)
if TYPE_CHECKING:
    Float = Any
else:
    Float = float


@dataclass
class PhysicalUnits:
    """Physical constants for units conversions."""

    BOLTZMANN_CONSTANT: float
    TO_ANGSTROM: float
    TO_PICO_SECOND: float
    TO_BAR: float
    TO_ELECTRON_VOLT: float
    TO_ATOMIC_MASS: float

    TO_NANO_METER: Float = _optional_attr
    TO_FEMTO_SECOND: Float = _optional_attr
    TO_NANO_SECOND: Float = _optional_attr
    TO_GIGA_PASCAL: Float = _optional_attr
    TO_PASCAL: Float = _optional_attr
    TO_ATMOSPHERE: Float = _optional_attr
    TO_KILO_BAR: Float = _optional_attr
    TO_KCAL_PER_MOL: Float = _optional_attr
    FROM_ANGSTROM: Float = _optional_attr
    FROM_NANO_METER: Float = _optional_attr
    FROM_FEMTO_SECOND: Float = _optional_attr
    FROM_PICO_SECOND: Float = _optional_attr
    FROM_NANO_SECOND: Float = _optional_attr
    FROM_GIGA_PASCAL: Float = _optional_attr
    FROM_KILO_BAR: Float = _optional_attr
    FROM_BAR: Float = _optional_attr
    FROM_ELECTRON_VOLT: Float = _optional_attr
    FROM_ATOMIC_MASS: Float = _optional_attr
    FROM_KCAL_PER_MOL: Float = _optional_attr

    def __post_init__(self) -> None:
        self.TO_NANO_METER = self.TO_ANGSTROM * 0.1
        self.TO_FEMTO_SECOND = self.TO_PICO_SECOND * 1000
        self.TO_NANO_SECOND = self.TO_PICO_SECOND * 0.001
        self.TO_GIGA_PASCAL = self.TO_BAR * 0.0001
        self.TO_PASCAL = self.TO_BAR * 100000
        self.TO_ATMOSPHERE = self.TO_BAR * 0.986923
        self.TO_KILO_BAR = self.TO_BAR * 0.001
        self.TO_KCAL_PER_MOL = self.TO_ELECTRON_VOLT * 23.0609
        self.FROM_ANGSTROM = 1 / self.TO_ANGSTROM
        self.FROM_NANO_METER = 1 / self.TO_NANO_METER
        self.FROM_FEMTO_SECOND = 1 / self.TO_FEMTO_SECOND
        self.FROM_PICO_SECOND = 1 / self.TO_PICO_SECOND
        self.FROM_NANO_SECOND = 1 / self.TO_NANO_SECOND
        self.FROM_GIGA_PASCAL = 1 / self.TO_GIGA_PASCAL
        self.FROM_KILO_BAR = 1 / self.TO_KILO_BAR
        self.FROM_BAR = 1 / self.TO_BAR
        self.FROM_ELECTRON_VOLT = 1 / self.TO_ELECTRON_VOLT
        self.FROM_ATOMIC_MASS = 1 / self.TO_ATOMIC_MASS
        self.FROM_KCAL_PER_MOL = 1 / self.TO_KCAL_PER_MOL


hartree_units = PhysicalUnits(
    BOLTZMANN_CONSTANT=3.166811563e-6,
    TO_ANGSTROM=5.29177249e-01,
    TO_PICO_SECOND=2.418884326e-05,
    TO_BAR=2.942102648e08,
    TO_ELECTRON_VOLT=2.7211407953e01,
    TO_ATOMIC_MASS=5.48579957163e-4,
)


units = hartree_units



from pathlib import Path
from typing import Optional

import jax
import jax.numpy as jnp
import numpy as np
from pantea.logger import logger
from pantea.types import Array, Dtype, default_dtype


class Scaler:
    """
    Scale descriptor values.

    Scaling parameters are calculated by fitting over the samples in the dataset.
    Available scaler information are as follows:

    * minimum
    * maximum
    * mean
    * sigma (standard deviation)

    Scaler is also used to warn when setting out-of-distribution samples base
    on the fitted scaler parameters.
    """

    def __init__(
        self,
        scale_type: str = "scale_center",
        scale_min: float = 0.0,
        scale_max: float = 1.0,
    ) -> None:
        """Initialize scaler including scaler type and min/max values."""
        assert scale_min < scale_max

        # Set min/max range for scaler
        self.scale_type: str = scale_type
        self.scale_min: float = scale_min
        self.scale_max: float = scale_max

        # Statistical parameters
        self.nsamples: int = 0  # number of samples
        self.dimension: int = 0  # dimension of each sample
        self.mean: Array = jnp.asarray(
            []
        )  # mean array of all fitted descriptor values
        self.sigma: Array = jnp.asarray([])  # standard deviation
        self.min: Array = jnp.asarray([])  # minimum
        self.max: Array = jnp.asarray([])  # maximum

        self.number_of_warnings: int = 0
        self.max_number_of_warnings: Optional[int] = None

        logger.debug(f"Initializing {self}")

        # Set scaler type function
        self._transform = getattr(self, f"{self.scale_type}")

    def fit(self, data: Array) -> None:
        """
        Fit scaler parameters using the given input data.
        Bach-wise sampling is also possible (see `this`_ for more details).

        .. _this: https://notmatthancock.github.io/2017/03/23/simple-batch-stat-updates.html
        """
        data = jnp.atleast_2d(data)  # type: ignore

        if self.nsamples == 0:
            self.nsamples = data.shape[0]
            self.dimension = data.shape[1]
            self.mean = jnp.mean(data, axis=0)
            self.sigma = jnp.std(data, axis=0)
            self.max = jnp.max(data, axis=0)
            self.min = jnp.min(data, axis=0)
        else:
            if data.shape[1] != self.dimension:
                logger.error(
                    f"Data dimension doesn't match: {data.shape[1]} (expected {self.dimension})",
                    exception=ValueError,
                )

            # New data (batch)
            new_mean: Array = jnp.mean(data, axis=0)
            new_sigma: Array = jnp.std(data, axis=0)
            new_min: Array = jnp.min(data, axis=0)
            new_max: Array = jnp.max(data, axis=0)
            m, n = self.nsamples, data.shape[0]

            # Calculate quantities for entire data
            mean = self.mean  # immutable
            self.mean = (
                m / (m + n) * mean + n / (m + n) * new_mean
            )  # self.mean is now a new array and different from the above mean variable
            self.sigma = jnp.sqrt(
                m / (m + n) * self.sigma**2
                + n / (m + n) * new_sigma**2
                + m * n / (m + n) ** 2 * (mean - new_mean) ** 2
            )
            self.max = jnp.maximum(self.max, new_max)
            self.min = jnp.minimum(self.min, new_min)
            self.nsamples += n

    def __call__(self, array: Array, warnings: bool = False) -> Array:
        """
        Transform the input descriptor values base on the scaler parameters.

        This method has to be called after fitting scaler over the dataset,
        or statistical parameters are already loaded (e.g. saved file).
        """
        if warnings:
            self._check_warnings(array)
        return self._transform(array)

    def set_max_number_of_warnings(self, number: Optional[int] = None) -> None:
        """Set the maximum number of warning for out of range descriptor values."""
        self.max_number_of_warnings = number
        self.number_of_warnings = 0
        logger.debug(
            f"Setting the maximum number of scaler warnings: {self.max_number_of_warnings}"
        )

    def _check_warnings(self, array: Array) -> None:
        """
        Check whether the output scaler values exceed the predefined min/max range values or not.

        If it's the case, it keeps counting the number of warnings and
        raises an error when it exceeds the maximum number.

        An out of range descriptor value is in fact an indication of
        the descriptor extrapolation which has to be avoided.
        """
        if self.max_number_of_warnings is None:
            return

        gt: Array
        lt: Array
        if array.ndim == 2:
            gt = jax.lax.gt(array, self.max[None, :])
            lt = jax.lax.gt(self.min[None, :], array)
        else:
            gt = jax.lax.gt(array, self.max)
            lt = jax.lax.gt(self.min, array)

        self.number_of_warnings += int(
            jnp.any(jnp.logical_or(gt, lt))
        )  # alternative counting is using sum

        if self.number_of_warnings >= self.max_number_of_warnings:
            logger.warning(
                "Exceeding maximum number scaler warnings (extrapolation warning): "
                f"{self.number_of_warnings} (max={self.max_number_of_warnings})"
            )

    def center(self, array: Array) -> Array:
        return array - self.mean

    def scale(self, array: Array) -> Array:
        return self.scale_min + (self.scale_max - self.scale_min) * (
            array - self.min
        ) / (self.max - self.min)

    def scale_center(self, array: Array) -> Array:
        return self.scale_min + (self.scale_max - self.scale_min) * (
            array - self.mean
        ) / (self.max - self.min)

    def scale_center_sigma(self, array: Array) -> Array:
        return (
            self.scale_min
            + (self.scale_max - self.scale_min)
            * (array - self.mean)
            / self.sigma
        )

    def save(self, filename: Path) -> None:
        """Save scaler parameters into file."""
        logger.debug(f"Saving scaler parameters into '{str(filename)}'")
        with open(str(filename), "w") as file:
            file.write(
                f"{'# Min':<23s} {'Max':<23s} {'Mean':<23s} {'Sigma':<23s}\n"
            )
            for i in range(self.dimension):
                file.write(
                    f"{self.min[i]:<23.15E} {self.max[i]:<23.15E} {self.mean[i]:<23.15E} {self.sigma[i]:<23.15E}\n"
                )

    def load(self, filename: Path, dtype: Optional[Dtype] = None) -> None:
        """Load scaler parameters from file."""
        logger.debug(f"Loading scaler parameters from '{str(filename)}'")
        data = np.loadtxt(str(filename)).T
        dtype = dtype if dtype is not None else default_dtype.FLOATX
        self.nsamples = 1
        self.dimension = data.shape[1]
        self.min = jnp.asarray(data[0], dtype=dtype)
        self.max = jnp.asarray(data[1], dtype=dtype)
        self.mean = jnp.asarray(data[2], dtype=dtype)
        self.sigma = jnp.asarray(data[3], dtype=dtype)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(scale_type='{self.scale_type}', "
            f"scale_min={self.scale_min}, scale_max={self.scale_max})"
        )

    def __bool__(self) -> bool:
        return False if len(self.mean) == 0 else True



from pantea.potentials.atomic_potential import AtomicPotential
from pantea.potentials.nnp.potential import NNP, NeuralNetworkPotential

__all__ = [
    "AtomicPotential",
    "NeuralNetworkPotential",
    "NNP",
]



from __future__ import annotations

import json
from pathlib import Path
from typing import Any, List

from pydantic import BaseModel


class _CFG(BaseModel):
    """A base class for default values as global configuration."""

    # Pydantic validators can be added to check attributes
    def __getitem__(self, keyword: str) -> Any:
        """Get value for the input name."""
        return getattr(self, keyword)

    def __setitem__(self, name, value) -> None:
        """Set value for the input name"""
        setattr(self, name, value)

    def keywords(self) -> List[str]:
        """Return a list of existing keyword names."""
        return self.__dict__.keys()  # type: ignore

    def to_json(self, file: Path) -> None:
        """Dump configuration into a json file."""
        with open(str(Path(file)), "w") as fp:
            json.dump(self.dict(), fp, indent=4)

    @classmethod
    def from_json(cls, file: Path) -> _CFG:
        """Create a configuration instance from the input json file."""
        with open(str(Path(file)), "r") as fp:
            kwargs = json.load(fp)
        return cls(**kwargs)



from dataclasses import replace
from typing import Optional, Protocol, Tuple

import jax.numpy as jnp
import numpy as np
from pantea.atoms.structure import Structure
from pantea.logger import logger
from pantea.types import Array, Element
from pantea.units import units

KB: float = units.BOLTZMANN_CONSTANT


class PotentialInterface(Protocol):
    def __call__(self, structure: Structure) -> Array:
        ...

    def compute_forces(self, structure: Structure) -> Array:
        ...


class MCSimulator:
    def __init__(
        self,
        potential: PotentialInterface,
        initial_structure: Structure,
        temperature: float,
        translate_step: float,
        movements_per_step: int = 1,
        atomic_masses: Optional[Array] = None,
        random_seed: int = 12345,
    ) -> None:
        """
        A Monte Carlo simulator to predict configuration space of particles.

        :param potential: input potential function
        :type potential: PotentialInterface
        :param initial_structure: initial structure of atoms
        :type initial_structure: Structure
        :param translate_step: maximum translate step in Bohr unit
        :type time_step: float
        :param temperature: desired temperature of the system
        :type temperature: float
        :param atomic_masses: atomic mass of atoms in the input structure, defaults to None
        :type atomic_masses: Optional[Array], optional
        :param random_seed: seed for generating random atom movements, defaults to 12345
        :type random_seed: int, optional

        This simulator is intended to be used for small systems of atoms during
        the potential training in order to generate new dataset.
        """

        logger.debug(f"Initializing {self.__class__.__name__}")
        self.potential = potential
        self.temperature = temperature
        self.translate_step = translate_step
        self.movements_per_step = movements_per_step

        self.masses: Array
        if atomic_masses is not None:
            self.masses = atomic_masses.reshape(-1, 1)
        else:
            logger.info("Extracting atomc masses from input structure")
            self.masses = initial_structure.get_masses().reshape(-1, 1)

        self.step: int = 0
        self.energy = self.potential(initial_structure)
        self._structure = replace(initial_structure)
        np.random.seed(random_seed)

    def update(self) -> None:
        """Update parameters for next time step."""
        self.metropolis_algorithm()
        self.step += 1

    def metropolis_algorithm(self) -> None:
        """Update atom positions based on metropolis algorithm."""
        displacements = np.random.uniform(
            low=-self.translate_step,
            high=self.translate_step,
            size=(self.movements_per_step, 3),
        )
        atom_indices = np.random.randint(
            low=0, high=self.natoms, size=(self.movements_per_step,)
        )
        new_position = self._structure.positions.at[atom_indices].add(
            displacements
        )
        new_structure = replace(self._structure, positions=new_position)
        new_energy = self.potential(new_structure)

        accept: bool = False
        if new_energy <= self.energy:
            accept = True
        else:
            prob = jnp.exp(
                -(new_energy - self.energy) / (KB * self.temperature)
            )
            accept = True if prob >= np.random.uniform(0.0, 1.0) else False

        if accept:
            self.energy = new_energy
            self._structure = new_structure

    def repr_physical_params(self) -> str:
        """Represent current physical parameters."""
        return (
            f"{self.step:<10} "
            f"Temp[K]:{self.temperature:<15.10f} "
            f"Epot[Ha]:{self.get_potential_energy():<15.10f} "
        )

    @property
    def positions(self) -> Array:
        return self._structure.positions

    @property
    def natoms(self) -> int:
        return self._structure.natoms

    def get_elements(self) -> Tuple[Element]:
        return self._structure.get_elements()

    def get_structure(self) -> Structure:
        return replace(
            self._structure,
            total_energy=self.potential(self._structure),
        )

    def get_potential_energy(self) -> Array:
        return self.potential(self._structure)



from typing import Any, Protocol

from pantea.types import Array


class DescriptorInterface(Protocol):
    """A base class for atomic environment descriptors."""

    def add(self, *args: Any, **kwargs: Any) -> None:
        ...

    def __call__(self, *args: Any, **kwargs: Any) -> Array:
        ...

    def grad(self, *args: Any, **kwargs: Any) -> Array:
        ...

    @property
    def num_descriptors(self) -> int:
        """Return number of items in the descriptor array."""
        ...

    @property
    def r_cutoff(self) -> float:
        """Return the cutoff radius of the descriptor."""
        ...



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass

import jax
import jax.numpy as jnp
from pantea.descriptors.acsf.cutoff import CutoffFunction
from pantea.descriptors.acsf.symmetry import BaseSymmetryFunction
from pantea.pytree import register_jax_pytree_node
from pantea.types import Array


class AngularSymmetryFunction(BaseSymmetryFunction, metaclass=ABCMeta):
    """A base class for `three body` (angular) symmetry functions."""

    def __hash__(self) -> int:
        """Enforce to use the parent class's hash method (JIT)."""
        return super().__hash__()

    @abstractmethod
    def __call__(
        self,
        rij: Array,
        rik: Array,
        rjk: Array,
        cost: Array,
    ) -> Array:
        ...


@dataclass
class G3(AngularSymmetryFunction):
    """Angular symmetry function."""

    cfn: CutoffFunction
    eta: float
    zeta: float
    lambda0: float
    r_shift: float

    def __hash__(self) -> int:
        """Enforce to use the parent class's hash method (JIT)."""
        return super().__hash__()

    @jax.jit
    def __call__(
        self,
        rij: Array,
        rik: Array,
        rjk: Array,
        cost: Array,
    ) -> Array:
        return (
            2.0 ** (1.0 - self.zeta)
            * jnp.power(1 + self.lambda0 * cost, self.zeta)
            * jnp.exp(-self.eta * (rij**2 + rik**2 + rjk**2))
            * self.cfn(rij)
            * self.cfn(rik)
            * self.cfn(rjk)
        )


@dataclass
class G9(G3):
    """
    Modified angular symmetry function.

    J. Behler, J. Chem. Phys. 134, 074106 (2011).
    """

    def __hash__(self) -> int:
        """Enforce to use the parent class's hash method (JIT)."""
        return super().__hash__()

    @jax.jit
    def __call__(
        self,
        rij: Array,
        rik: Array,
        rjk: Array,
        cost: Array,
    ) -> Array:
        # TODO: r_shift, define params argument instead
        return (
            2.0 ** (1.0 - self.zeta)
            * jnp.power(1 + self.lambda0 * cost, self.zeta)
            * jnp.exp(-self.eta * (rij**2 + rik**2))
            * self.cfn(rij)
            * self.cfn(rik)
        )


register_jax_pytree_node(G3)
register_jax_pytree_node(G9)



from pathlib import Path
from typing import Optional, Protocol

import ase
import ase.io
from pantea.atoms.structure import Structure
from pantea.logger import logger


class SimulatorInterface(Protocol):
    step: int

    def repr_physical_params(self) -> None:
        ...

    def get_structure(self) -> Structure:
        ...

    def update(self) -> None:
        ...


def run_simulation(
    simulator: SimulatorInterface,
    num_steps: int = 1,
    output_freq: Optional[int] = None,
    filename: Optional[Path] = None,
    append: bool = False,
) -> None:
    """
    Run an input simulator for a given number of steps.

    :param simulator: either a molecular dynamics (MD) or monte carlo (MC) simulator
    :type simulator: SimulatorInterface
    :param num_steps: number of steps, defaults to 1
    :type num_steps: int, optional
    :param output_freq: print outputs and/or dump configuration file after this number of steps, defaults to None
    :type output_freq: Optional[int], optional
    :param filename: output configuration file (e.g., `*.xyz`) as supported in the ASE, defaults to None
    :type filename: Optional[Path], optional
    :param append: whether append to the exiting configuration file or not, defaults to False
    :type append: bool, optional


    The input simulator, for example, can be molecular dynamics (MD) or
    monte carlo (MC).
    """

    cls_name = simulator.__class__.__name__
    logger.info(f"Running {cls_name} for {num_steps} steps")

    if output_freq is None:
        output_freq = 1 if num_steps < 100 else int(0.01 * num_steps)
    is_output: bool = output_freq > 0

    if filename is not None:
        filename = Path(filename)
        if is_output:
            logger.info(
                f"Writing atomic configuration into '{filename.name}' "
                f"file every {output_freq} steps"
            )
        if not append:
            with open(str(filename), "w"):
                pass

    init_step: int = simulator.step
    try:
        for _ in range(num_steps):
            if is_output and ((simulator.step - init_step) % output_freq == 0):
                print(simulator.repr_physical_params())
                if filename is not None:
                    atoms = simulator.get_structure().to_ase()
                    ase.io.write(str(filename), atoms, append=True)
            simulator.update()
    except KeyboardInterrupt:
        print("KeyboardInterrupt")
    if is_output:
        print(simulator.repr_physical_params())



from typing import Dict, List, Mapping, Set, Tuple, Union

from pantea.logger import logger
from pantea.types import Element
from pantea.units import units

# fmt: off
_KNOWN_ELEMENTS_LIST: Tuple = (
    "H", "He", "Li", "Be", "B", "C", "N", "O", "F", "Ne",
    "Na", "Mg", "Al", "Si", "P", "S", "Cl", "Ar", "K", "Ca",
    "Sc", "Ti", "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn",
    "Ga", "Ge", "As", "Se", "Br", "Kr", "Rb", "Sr", "Y", "Zr",
    "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn",
    "Sb", "Te", "I", "Xe", "Cs", "Ba", "La", "Ce", "Pr", "Nd",
    "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb",
    "Lu", "Hf", "Ta", "W", "Re", "Os", "Ir", "Pt", "Au", "Hg",
    "Tl", "Pb", "Bi", "Po", "At", "Rn", "Fr", "Ra", "Ac", "Th",
    "Pa", "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm",
    "Md",
)

_KNOWN_ELEMENTS_DICT_MASS = {
    'H': 1.008, 'He': 4.003, 'Li': 6.941, 'Be': 9.012,
    'B': 10.811, 'C': 12.011, 'N': 14.007, 'O': 15.999,
    'F': 18.998, 'Ne': 20.180, 'Na': 22.990, 'Mg': 24.305,
    'Al': 26.982, 'Si': 28.086, 'P': 30.974, 'S': 32.066,
    'Cl': 35.453, 'Ar': 39.948, 'K': 39.098, 'Ca': 40.078,
    'Sc': 44.956, 'Ti': 47.867, 'V': 50.942, 'Cr': 51.996,
    'Mn': 54.938, 'Fe': 55.845, 'Co': 58.933, 'Ni': 58.693,
    'Cu': 63.546, 'Zn': 65.38, 'Ga': 69.723, 'Ge': 72.631,
    'As': 74.922, 'Se': 78.971, 'Br': 79.904, 'Kr': 84.798,
    'Rb': 84.468, 'Sr': 87.62, 'Y': 88.906, 'Zr': 91.224,
    'Nb': 92.906, 'Mo': 95.95, 'Tc': 98.907, 'Ru': 101.07,
    'Rh': 102.906, 'Pd': 106.42, 'Ag': 107.868, 'Cd': 112.414,
    'In': 114.818, 'Sn': 118.711, 'Sb': 121.760, 'Te': 126.7,
    'I': 126.904, 'Xe': 131.294, 'Cs': 132.905, 'Ba': 137.328,
    'La': 138.905, 'Ce': 140.116, 'Pr': 140.908, 'Nd': 144.243,
    'Pm': 144.913, 'Sm': 150.36, 'Eu': 151.964, 'Gd': 157.25,
    'Tb': 158.925, 'Dy': 162.500, 'Ho': 164.930, 'Er': 167.259,
    'Tm': 168.934, 'Yb': 173.055, 'Lu': 174.967, 'Hf': 178.49,
    'Ta': 180.948, 'W': 183.84, 'Re': 186.207, 'Os': 190.23,
    'Ir': 192.217, 'Pt': 195.085, 'Au': 196.967, 'Hg': 200.592,
    'Tl': 204.383, 'Pb': 207.2, 'Bi': 208.980, 'Po': 208.982,
    'At': 209.987, 'Rn': 222.081, 'Fr': 223.020, 'Ra': 226.025,
    'Ac': 227.028, 'Th': 232.038, 'Pa': 231.036, 'U': 238.029,
    'Np': 237, 'Pu': 244, 'An': 243, 'Cm': 247, 'Bk': 247,
    'Ct': 251, 'Es': 252, 'Fm': 257, 'Mf': 258, 'No': 259,
    'Lr': 262, 'Rf': 261, 'Db': 262, 'Sg': 266, 'Bh': 264,
    'Hs': 269, 'Mt': 268, 'Ds': 271, 'Rg': 272, 'Cn': 285,
    'Nh': 284, 'Fl': 289, 'Mc': 288, 'Lv': 292, 'Ts': 294,
    'Og': 294,
}
# fmt: on


_KNOWN_ELEMENTS_DICT: Mapping[Element, int] = {
    element: atomic_number
    for atomic_number, element in enumerate(_KNOWN_ELEMENTS_LIST, start=1)
}


class ElementMap:
    """
    Mapping element to atomic number utility class.

    It assigns an atomic type to each element which allows
    efficient array processing (e.g. applying conditions) using
    the arrays of integer (i.e. atom types) instead of strings.
    """

    def __init__(self, elements: List[Element]) -> None:
        """Initialize element map."""
        self.unique_elements: Set[Element] = set(elements)
        self._element_to_atomic_number: Dict[Element, int] = dict()
        self._element_to_atom_type: Dict[Element, int] = dict()
        self._atom_type_to_element: Dict[int, Element] = dict()
        self._create_mapping_dicts()
        logger.debug(f"Initialized {self.__class__.__name__}")

    def __call__(self, item: Union[Element, int]) -> Union[int, Element]:
        """Map an element to the atom type and vice versa."""
        result: Union[int, Element]
        if isinstance(item, int):
            return self._atom_type_to_element[item]
        elif isinstance(item, Element):
            return self._element_to_atom_type[item]
        else:
            logger.error(
                f"Unknown item type '{type(item)}'", exception=TypeError
            )
            return  # type: ignore

    def _create_mapping_dicts(self) -> None:
        """
        Create dictionary to map elements, atom types, and atomic numbers.
        The atom types are sorted based on elements' atomic number.
        """
        self._element_to_atomic_number = {
            elem: _KNOWN_ELEMENTS_DICT[elem] for elem in self.unique_elements
        }
        self._element_to_atom_type = {
            elem: atom_type
            for atom_type, elem in enumerate(
                sorted(
                    self._element_to_atomic_number,
                    key=self._element_to_atomic_number.get,  # type: ignore
                ),
                start=1,
            )
        }
        self._atom_type_to_element = {
            atom_type: elem
            for elem, atom_type in self._element_to_atom_type.items()
        }

    @classmethod
    def get_atomic_number(cls, element: Element) -> int:
        """
        Return atomic number of the input element.

        :param element: element name
        :return: atomic number
        """
        return _KNOWN_ELEMENTS_DICT[element]

    @classmethod
    def get_element(cls, atomic_number: int) -> Element:
        """
        Return element name of the given atomic number.

        :param atomic_number: atomic number
        :return: element name
        """
        return _KNOWN_ELEMENTS_LIST[atomic_number - 1]

    @property
    def atom_type_to_element(self) -> Dict[int, Element]:
        """
        Return a dictionary mapping of atom type to element.

        This property is defined due to a serialization issue
        of the ElementMap class during parallelization.

        :return: a dictionary of mapping an atom type (integer)
        to the corresponding element (string)
        """
        return self._atom_type_to_element

    @property
    def element_to_atom_type(self) -> Dict[Element, int]:
        """
        Return a mapping dictionary of element to atom type.

        This property is defined due to a serialization issue of
        the ElementMap class during parallelization.

        :return: a dictionary of mapping an element (string) to
        the corresponding atom type (integer)
        """
        return self._element_to_atom_type

    @classmethod
    def atomic_number_to_element(cls, atomic_number: int) -> Element:
        return _KNOWN_ELEMENTS_LIST[atomic_number - 1]

    @classmethod
    def element_to_atomic_number(cls, element: Element) -> int:
        return _KNOWN_ELEMENTS_DICT[element]

    @classmethod
    def element_to_atomic_mass(cls, element: Element) -> float:
        return _KNOWN_ELEMENTS_DICT_MASS[element] * units.FROM_ATOMIC_MASS



from pantea.simulation.mc import MCSimulator
from pantea.simulation.md import MDSimulator
from pantea.simulation.run import run_simulation
from pantea.simulation.thermostat import BrendsenThermostat

__all__ = [
    "MDSimulator",
    "BrendsenThermostat",
    "MCSimulator",
    "run_simulation",
]



from pantea.atoms.box import Box
from pantea.atoms.element import Element
from pantea.atoms.neighbor import Neighbor
from pantea.atoms.structure import Structure

__all__ = [
    "Structure",
    "Box",
    "Element",
    "Neighbor",
]



from typing import List, Optional, Tuple


def tokenize(
    line: str,
    comment: Optional[str] = None,
) -> Tuple[Optional[str], List[str]]:
    """An utility function to read the input line as a keyword and list of tokens."""
    # Skip comments
    if comment is not None:
        if line.startswith(comment):
            return (None, list())
        else:
            line = line[: line.find(comment)]
    # Find keyword and values
    tokens: List[str] = line.rstrip("/n").split()
    if len(tokens) > 1:
        return (tokens[0].lower(), tokens[1:])
    elif len(tokens) > 0:
        return (tokens[0].lower(), list())
    else:
        return (None, list())



import os
from pathlib import Path
from typing import Tuple

os.environ["JAX_ENABLE_X64"] = "1"
os.environ["JAX_PLATFORM_NAME"] = "cpu"

import jax.numpy as jnp
import pytest
from pantea.datasets import RunnerDataset
from pantea.potentials import NeuralNetworkPotential
from pantea.potentials.nnp.settings import (
    NeuralNetworkPotentialSettings as PotentialSettings,
)

dataset_file = Path("tests", "h2o.data")
potential_file = Path("tests", "h2o.json")


class TestNeuralNetworkPotential:
    dataset = RunnerDataset(dataset_file)
    nnp: NeuralNetworkPotential = NeuralNetworkPotential.from_file(
        potential_file
    )

    @pytest.mark.parametrize(
        "nnp, expected",
        [
            (
                nnp,
                (
                    2,
                    ("H", "O"),
                ),
            ),
        ],
    )
    def test_settings(
        self,
        nnp: NeuralNetworkPotential,
        expected: Tuple,
    ) -> None:
        assert nnp.num_elements == expected[0]
        assert nnp.elements == expected[1]
        assert nnp.settings == PotentialSettings.from_json(potential_file)

    @pytest.mark.parametrize(
        "nnp, dataset, expected",
        [
            (
                nnp,
                dataset,
                (
                    jnp.asarray(-0.00721363),
                    jnp.asarray(
                        [
                            [-0.00343415, 0.00153666, -0.0203776],
                            [-0.11087799, -0.04327171, 0.10297609],
                            [-0.01472104, 0.0449558, 0.04925348],
                            [-0.02519826, 0.00152729, -0.01613272],
                            [0.06494806, -0.02886784, -0.01245033],
                            [-0.04200673, -0.00515676, -0.07121348],
                            [0.03067259, 0.01101292, -0.04297253],
                            [0.09775139, 0.06042628, -0.10140687],
                            [-0.02604892, 0.04006792, -0.13405925],
                            [0.01066793, -0.00437668, 0.03239093],
                            [0.05837363, 0.07088667, -0.09498966],
                            [0.04761543, -0.03721519, -0.02089787],
                        ]
                    ),
                ),
            ),
        ],
    )
    def test_outputs(
        self,
        nnp: NeuralNetworkPotential,
        dataset: RunnerDataset,
        expected: Tuple,
    ) -> None:
        nnp.fit_scaler(dataset)

        # total energy
        assert jnp.allclose(nnp(dataset[0]), expected[0])
        assert jnp.allclose(nnp.compute_forces(dataset[0]), expected[1])

    @pytest.mark.parametrize(
        "nnp, dataset",
        [
            (
                nnp,
                dataset,
            ),
        ],
    )
    def test_save_and_load(
        self,
        nnp: NeuralNetworkPotential,
        dataset: RunnerDataset,
    ) -> None:
        structure = dataset[0]
        nnp.output_dir = potential_file.parent
        nnp.save()

        settings_new = nnp.settings.copy()
        settings_new.random_seed = 4321
        nnp_new = NeuralNetworkPotential(settings=settings_new)
        nnp_new.fit_scaler(dataset)
        nnp_new.output_dir = potential_file.parent

        assert not nnp.settings == nnp_new.settings
        assert not jnp.allclose(nnp(structure), nnp_new(structure))
        nnp_new.load()
        assert jnp.allclose(nnp(structure), nnp_new(structure))



from abc import ABCMeta, abstractmethod
from typing import NamedTuple, Optional

from pantea.descriptors.acsf.cutoff import CutoffFunction
from pantea.logger import logger
from pantea.pytree import BaseJaxPytreeDataClass
from pantea.types import Array, Element


class BaseSymmetryFunction(BaseJaxPytreeDataClass, metaclass=ABCMeta):
    """
    A base class for symmetry functions.
    All symmetry functions (i.e. radial and angular) must derive from this base class.
    """

    def __init__(self, cfn: CutoffFunction) -> None:
        self.cfn: CutoffFunction = cfn
        logger.debug(repr(self))

    @abstractmethod
    def __call__(self, *args, **kwargs) -> Array:
        ...

    @property
    def r_cutoff(self) -> float:
        return self.cfn.r_cutoff


class EnvironmentElements(NamedTuple):
    """
    Representative elements for the chemical environment including central
    elements and its neighbors.
    """

    central: Element
    neighbor_j: Element
    neighbor_k: Optional[Element] = None



import os
from typing import Tuple

import jax.numpy as jnp
import pytest
from ase import Atoms
from pantea.atoms.structure import Structure
from pantea.simulation.lj import LJPotential
from pantea.simulation.md import MDSimulator
from pantea.units import units

os.environ["JAX_ENABLE_X64"] = "1"
os.environ["JAX_PLATFORM_NAME"] = "cpu"


def get_structure() -> Structure:
    d = 6  # Angstrom
    uc = Atoms("He", positions=[(d / 2, d / 2, d / 2)], cell=(d, d, d))
    s = Structure.from_ase(uc.repeat((2, 2, 2)))
    return s


def get_potential() -> LJPotential:
    return LJPotential(
        sigma=2.5238 * units.FROM_ANGSTROM,
        epsilon=4.7093e-04 * units.FROM_ELECTRON_VOLT,
        r_cutoff=6.3095 * units.FROM_ANGSTROM,
    )


class TestMDSimulator:
    md = MDSimulator(
        potential=get_potential(),
        initial_structure=get_structure(),
        time_step=0.5 * units.FROM_FEMTO_SECOND,
        temperature=300.0,
    )

    @pytest.mark.parametrize(
        "md, expected",
        [
            (
                md,
                (
                    0.5 * units.FROM_FEMTO_SECOND,
                    262.15531167,
                    5.70328776e-07,
                    0.00995778,
                    jnp.asarray([11.33835602, 11.33835602, 11.33835602]),
                ),
            ),
        ],
    )
    def test_general_attributes(
        self,
        md: MDSimulator,
        expected: Tuple,
    ) -> None:
        assert md.step == 0
        assert jnp.allclose(md.elapsed_time, 0.0)
        assert jnp.allclose(md.time_step, expected[0])
        assert jnp.allclose(md.get_center_of_mass_velocity(), jnp.zeros(3))
        assert jnp.allclose(md.temperature, expected[1])
        assert jnp.allclose(md.get_pressure(), expected[2])
        assert jnp.allclose(md.get_total_energy(), expected[3])
        assert jnp.allclose(md.get_center_of_mass_position(), expected[4])

    @pytest.mark.parametrize(
        "md, structure",
        [
            (
                md,
                get_structure(),
            ),
        ],
    )
    def test_structure_attributes(
        self,
        md: MDSimulator,
        structure: Structure,
    ) -> None:
        assert jnp.allclose(md.positions, structure.positions)
        assert jnp.allclose(md.masses, structure.get_masses())

    @pytest.mark.parametrize(
        "md, expected",
        [
            (
                md,
                (
                    0.5 * units.FROM_FEMTO_SECOND,
                    261.23330876,
                    5.68325481e-07,
                    0.00992273,
                    jnp.asarray([11.33835602, 11.33835602, 11.33835602]),
                ),
            ),
        ],
    )
    def test_update(
        self,
        md: MDSimulator,
        expected: Tuple,
    ) -> None:
        md.update()
        assert md.step == 1
        assert jnp.allclose(md.elapsed_time, expected[0])
        assert jnp.allclose(md.time_step, expected[0])
        assert jnp.allclose(md.get_center_of_mass_velocity(), jnp.zeros(3))
        assert jnp.allclose(md.get_center_of_mass_position(), expected[4])



import os
from typing import Tuple

os.environ["JAX_ENABLE_X64"] = "1"
os.environ["JAX_PLATFORM_NAME"] = "cpu"

import jax.numpy as jnp
import pytest
from pantea.atoms.structure import Structure
from pantea.descriptors.acsf import ACSF, G2, G3, CutoffFunction
from pantea.descriptors.acsf.symmetry import EnvironmentElements
from pantea.types import Array


# @pytest.fixture
def lj_acsf() -> ACSF:
    """Initialize using the `add` method."""
    acsf: ACSF = ACSF("Ne")
    cfn: CutoffFunction = CutoffFunction.from_cutoff_type(
        r_cutoff=3.0, cutoff_type="tanhu"
    )
    acsf.add(G2(cfn, eta=1.00, r_shift=0.00), "Ne")
    acsf.add(G2(cfn, eta=1.00, r_shift=0.25), "Ne")
    acsf.add(G2(cfn, eta=1.00, r_shift=0.50), "Ne")
    acsf.add(G2(cfn, eta=1.00, r_shift=0.75), "Ne")
    acsf.add(G2(cfn, eta=1.00, r_shift=1.00), "Ne")
    return acsf


def h2o_acsf() -> ACSF:
    """Initialize directly from the radial and angular terms."""
    return ACSF(
        central_element="O",
        radial_symmetry_functions=(
            (
                EnvironmentElements(central="O", neighbor_j="H"),
                G2(
                    cfn=CutoffFunction.from_cutoff_type(
                        r_cutoff=5.9043202, cutoff_type="tanhu"
                    ),  # r_cutoff = box.length / 2
                    r_shift=0.0,
                    eta=0.001,
                ),
            ),
        ),
        angular_symmetry_functions=(
            (
                EnvironmentElements(
                    central="O", neighbor_j="H", neighbor_k="H"
                ),
                G3(
                    cfn=CutoffFunction.from_cutoff_type(
                        r_cutoff=5.9043202, cutoff_type="tanhu"
                    ),  # r_cutoff = box.length / 2
                    eta=0.07,
                    zeta=1.0,
                    lambda0=1.0,
                    r_shift=0.0,
                ),
            ),
        ),
    )


class TestACSF:
    lj_structure: Structure = Structure.from_dict(
        {
            "positions": [
                [0.0, 0.0, 0.0],
                [0.588897275, 0.588897275, 0.588897275],
            ],
            "elements": ["Ne", "Ne"],
        }
    )
    h2o_structure: Structure = Structure.from_dict(
        {
            "lattice": [
                [11.8086403654, 0.0, 0.0],
                [0.0, 11.8086403654, 0.0],
                [0.0, 0.0, 11.8086403654],
            ],
            "positions": [
                [0.2958498542, -0.8444146738, 1.9618569793],
                [1.7226932399, -1.8170359274, 0.5237867306],
                [1.2660050151, -4.3958431356, 0.822408813],
                [2.5830855482, 6.6971705174, 2.042321518],
                [6.3733092527, 4.0651032678, 1.6571254123],
                [4.5893132971, 4.2984466507, 3.33215409],
                [-2.2136818837, 1.3673642557, 2.7768013741],
                [-1.9466635812, -0.3397746484, 3.3828743394],
                [-0.4559776878, 0.1681476423, -5.430449296],
                [-1.2831542798, 4.5473991714, -2.294184217],
                [-2.3516507887, 0.9376745482, -4.0756290424],
                [-1.7819663898, 4.1957022409, -4.0621741923],
            ],
            "elements": [
                "O",
                "H",
                "H",
                "O",
                "H",
                "H",
                "O",
                "H",
                "H",
                "O",
                "H",
                "H",
            ],
        }
    )

    @pytest.mark.parametrize(
        "acsf, expected",
        [
            (
                lj_acsf(),
                ("Ne", 5, 0, 5, 3.0),
            ),
        ],
    )
    def test_acsf_attributes(
        self,
        acsf: ACSF,
        expected: Tuple,
    ) -> None:
        assert acsf.central_element == expected[0]
        assert acsf.num_radial_symmetry_functions == expected[1]
        assert acsf.num_angular_symmetry_functions == expected[2]
        assert acsf.num_symmetry_functions == expected[3]
        assert acsf.r_cutoff == expected[4]

    @pytest.mark.parametrize(
        "acsf, structure, expected",
        [
            (
                lj_acsf(),
                lj_structure,
                jnp.tile(
                    jnp.asarray(
                        [
                            0.0683537673,
                            0.1069323809,
                            0.1476281601,
                            0.1798632214,
                            0.1933876418,
                        ]
                    ),
                    (2, 1),
                ),
            ),
        ],
    )
    def test_acsf_without_pbc(
        self,
        acsf: ACSF,
        structure: Structure,
        expected: Array,
    ) -> None:
        assert acsf(structure).shape == expected.shape
        assert jnp.allclose(acsf(structure), expected)

    @pytest.mark.parametrize(
        "acsf, structure, expected",
        [
            (
                h2o_acsf(),
                h2o_structure,
                (
                    (4, 2),
                    jnp.asarray([[0.3333029149, 0.0002204830]]),
                ),
            ),
        ],
    )
    def test_acsf_with_pbc(
        self,
        acsf: ACSF,
        structure: Structure,
        expected: Array,
    ) -> None:
        assert acsf(structure).shape == expected[0]
        assert jnp.allclose(
            acsf(structure, atom_indices=jnp.asarray(0)), expected[1]
        )



from typing import Tuple

from flax import linen as nn
from jax.random import KeyArray
from pantea.types import Array, Dtype


class UniformInitializer:
    """Custom uniform initializer for the FLAX model."""

    def __init__(self, weights_range: Tuple[float, float]) -> None:
        self.weights_range = weights_range
        self.initializer = nn.initializers.uniform(
            self.weights_range[1] - self.weights_range[0]
        )

    def __call__(
        self, rng: KeyArray, shape: Tuple[int, ...], dtype: Dtype
    ) -> Array:
        return self.initializer(rng, shape, dtype) + self.weights_range[0]



import logging
import sys
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Any


class Logger:
    """
    An utility logger class which allows more tweaks than the standard python logging.
    """

    def __init__(self, level=logging.WARNING, filename=None) -> None:
        self.logger: Logger = logging.getLogger("Pantea")  # type: ignore
        self.level: int = level
        self.handlers = list()

        self._add_console_handler()
        if filename:
            self._add_file_handler(filename)

        for handler in logging.root.handlers:
            logging.root.removeHandler(handler)
        logging.basicConfig(level=logging.DEBUG, handlers=self.handlers)

        # Avoiding matplotlib debug messages
        logging.getLogger("matplotlib").setLevel(logging.ERROR)
        logging.getLogger("matplotlib.font_manager").disabled = True
        logging.getLogger("jax").setLevel(logging.WARNING)

    def set_level(self, level) -> None:
        """
        Set the console logging level.

        :param level: INFO, WARNING, ERROR, and DEBUG
        """
        self.level = level
        logging.root.handlers[0].setLevel(self.level)

    def _add_console_handler(self) -> None:
        """
        Prepare console handler and adding it as a default handler.
        """
        console_handler = logging.StreamHandler(sys.stderr)
        console_handler.setFormatter(
            logging.Formatter("%(levelname)s %(message)s")
        )  # logging.Formatter('[%(asctime)s] %(levelname)s: %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')
        console_handler.setLevel(self.level)
        self.handlers.append(console_handler)

    def _add_file_handler(self, filename: Path) -> None:
        """
        Add a file handler to the logging.

        :param filename: output log file name
        :type filename: Path
        """
        # File handler is also useful for multi-process logging.
        file_handler = RotatingFileHandler(
            str(Path(filename)), "a", 1_000_000, 3
        )
        file_handler.setFormatter(
            logging.Formatter(
                "%(levelname)-8s %(message)s", datefmt="%m/%d/%Y %I:%M:%S %p"
            )
        )
        file_handler.setLevel(logging.DEBUG)
        self.handlers.append(file_handler)

    def debug(self, msg, *args, **kwargs) -> None:
        self.logger.debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs) -> None:
        self.logger.info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs) -> None:
        self.logger.warning(msg, *args, **kwargs)

    def warn(self, msg, *args, **kwargs) -> None:
        self.warning(msg, *args, **kwargs)

    def error(self, msg, exception: Any = None, *args, **kwargs) -> None:
        self.logger.error(msg, *args, **kwargs)
        if exception is not None:
            raise exception(msg)


# Create the global logger object
logger = Logger()  # filename="debug.log"


def set_logging_level(level: int) -> None:
    """
    Set the global logging level

    :param level: INFO, WARNING, ERROR, and DEBUG
    :type level: int (logging)
    """
    logger.set_level(level)


class LoggingContextManager:
    """
    This utility class allows to monitor logging messages with different level
    (e.g. INFO or DEBUG) for a particular section of the scripts.
    """

    def __init__(self, level) -> None:
        self.level = level
        self.current_level = None

    def __enter__(self) -> Logger:
        global logger
        self.current_level = logger.level
        logger.set_level(self.level)
        return logger

    def __exit__(self, type, value, traceback) -> None:
        global logger
        logger.set_level(self.current_level)



from typing import Dict, Protocol, Tuple

from frozendict import frozendict
from pantea.potentials.atomic_potential import AtomicPotential
from pantea.potentials.nnp.settings import (
    NeuralNetworkPotentialSettings as PotentialSettings,
)
from pantea.types import Element


class NeuralNetworkPotentialInterface(Protocol):
    """Interface for neural network potential (NNP)."""

    settings: PotentialSettings
    elements: Tuple[Element, ...]
    atomic_potential: Dict[Element, AtomicPotential]
    model_params: Dict[Element, frozendict]



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import (
    Any,
    DefaultDict,
    Dict,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Tuple,
)

import jax
import jax.numpy as jnp
import numpy as np
from ase import Atoms as AseAtoms
from jax import tree_util

from pantea.atoms._structure import (
    _calculate_center_of_mass,
    _calculate_distances,
)
from pantea.atoms.box import Box
from pantea.atoms.element import ElementMap
from pantea.atoms.neighbor import Neighbor
from pantea.logger import logger
from pantea.pytree import BaseJaxPytreeDataClass, register_jax_pytree_node
from pantea.types import Array, Dtype, Element, default_dtype
from pantea.units import units


@dataclass
class Structure(BaseJaxPytreeDataClass):
    """
    A structure in the context of simulation box consists of
    arrays that store atomic attributes for a collection of atoms.

    The attributes of atoms within a structure can be described as follows:

    * `positions`: position of atoms
    * `forces`: force components
    * `energies`: associated atom energies
    * `total_energy`: the total energy of atoms
    * `charges`:  electric charges
    * `total_charge`: the total charge of atoms

    The structure serves as a fundamental data unit for the atoms in the simulation box.
    Multiple structures can be gathered into a list to train a potential, or alternatively,
    the total energy and force components can be computed for a specific structure.

    Each structure has three additional instances:

    * `Box`: applying periodic boundary condition (PBC) along x, y, and z directions
    * `ElementMap`: determines how to extract assigned atom types from the element and vice versa
    * `Neighbor`: computes the list of neighboring atoms inside a specified cutoff radius

    .. note::
        The structure can be viewed as a separate domain for implementing MPI in large-scale
        molecular dynamics (MD) simulations, as demonstrated in the `miniMD`_ program.

    .. _miniMD: https://github.com/Mantevo/miniMD
    """

    positions: Array
    forces: Array
    energies: Array
    total_energy: Array
    charges: Array
    total_charge: Array
    atom_types: Array
    element_map: ElementMap
    box: Optional[Box] = None
    neighbor: Optional[Neighbor] = None

    def __post_init__(self) -> None:
        logger.debug(f"Initializing {self.__class__.__name__}()")
        self._assert_jit_dynamic_attributes(
            expected=(
                "positions",
                "forces",
                "energies",
                "total_energy",
                "charges",
                "total_charge",
                "atom_types",
            )
        )
        self._assert_jit_static_attributes(
            expected=(
                "element_map",
                "box",
                "neighbor",
            )
        )
        if self.box is not None:
            self.positions = self.box.shift_inside_box(self.positions)

    @classmethod
    def from_dict(
        cls,
        data: Dict[str, Any],
        dtype: Optional[Dtype] = None,
    ) -> Structure:
        """
        Instantiate a structure object using an input data dictionary that contains
        distinct lists of positions, forces, elements, lattice, etc.

        :param data: input data
        :param r_cutoff: neighbor atoms cutoff radius, defaults to None
        :param dtype: data type for arrays, defaults to None
        :return: the initialized Structure
        """
        if dtype is None:
            dtype = default_dtype.FLOATX

        input_data: DefaultDict[str, List] = defaultdict(list, data)
        inputs: Dict[str, Any] = dict()
        try:
            element_map: ElementMap = ElementMap(input_data["elements"])
            inputs.update(
                cls._init_arrays(
                    input_data,
                    element_map=element_map,
                    dtype=dtype,
                ),
            )
            inputs["element_map"] = element_map
            inputs["box"] = cls._init_box(input_data["lattice"], dtype=dtype)
        except KeyError:
            logger.error(
                "Cannot find at least one of the expected keyword in the input data.",
                exception=KeyError,
            )
        return cls(**inputs)

    @classmethod
    def from_ase(
        cls,
        atoms: AseAtoms,
        dtype: Optional[Dtype] = None,
    ) -> Structure:
        """
        Create an instance of the structure based on the input `ASE`_ atoms.

        :param atoms: input `ASE`_ atoms instance
        :param r_cutoff: neighbor atoms cutoff radius, defaults to None
        :param dtype: data type for arrays, defaults to None
        :return: the initialized structure

        .. _ASE: https://wiki.fysik.dtu.dk/ase/index.html
        """
        if dtype is None:
            dtype = default_dtype.FLOATX

        inputs: Dict[str, Any] = dict()
        data = {
            "elements": [
                ElementMap.atomic_number_to_element(n)
                for n in atoms.get_atomic_numbers()
            ],
            "lattice": np.array(
                atoms.get_cell() * units.FROM_ANGSTROM, dtype=dtype
            ),
            "positions": atoms.get_positions() * units.FROM_ANGSTROM,
        }
        for attr, ase_attr in zip(
            ("charges", "energies"),
            ("charges", "potential_energies"),
        ):
            try:
                data[attr] = getattr(atoms, f"get_{ase_attr}")()
            except RuntimeError:
                continue
        for attr in ("energies", "charges"):
            if attr in data:
                data[f"total_{attr}"] = sum(data[attr])

        input_data: DefaultDict[str, List] = defaultdict(list, data)
        try:
            element_map: ElementMap = ElementMap(input_data["elements"])
            inputs.update(
                cls._init_arrays(
                    input_data, element_map=element_map, dtype=dtype
                ),
            )
            inputs["element_map"] = element_map
            inputs["box"] = cls._init_box(input_data["lattice"], dtype=dtype)
        except KeyError:
            logger.error(
                "Can not find at least one of the expected keyword in the input data.",
                exception=KeyError,
            )
        return cls(**inputs)

    @classmethod
    def _init_arrays(
        cls,
        data: Dict[str, Any],
        element_map: ElementMap,
        dtype: Dtype,
    ) -> Dict[str, Array]:
        """Initialize array atomic attributes from the input data dictionary."""
        logger.debug(f"{cls.__name__} is allocating arrays as follows:")
        arrays: Dict[str, Array] = dict()
        for atom_attr in Structure._get_atom_attributes():
            try:
                array: Array
                if atom_attr == "atom_types":
                    array = jnp.array(
                        [element_map(atom) for atom in data["elements"]],
                        dtype=default_dtype.INDEX,
                    )
                else:
                    array = jnp.array(data[atom_attr], dtype=dtype)
                arrays[atom_attr] = array
                logger.debug(
                    f"{atom_attr:12} -> Array(shape={array.shape}, dtype='{array.dtype}')"
                )
            except KeyError:
                logger.error(
                    f"Cannot find atom attribute {atom_attr} in the input data",
                    exception=KeyError,
                )
        return arrays

    @classmethod
    def _init_box(
        cls,
        lattice: List[float],
        dtype: Dtype,
    ) -> Optional[Box]:
        """Initialize simulation box from input lattice matrix."""
        if len(lattice) > 0:
            return Box.from_list(lattice, dtype=dtype)
        else:
            logger.debug("No lattice info were found")

    def update_neighbor(self, r_cutoff: Optional[float] = None) -> None:
        """
        Update the neighbor list, building it if required.


        This is useful for efficiently determining the neighboring atoms within
        a specified cutoff radius. The neighbor list allows for faster calculations
        properties that depend on nearby atoms, such as computing forces, energies,
        or evaluating interatomic distances.

        There are various scenarios that may necessitate updating the neighbor list,
        including changes in the positions of atoms within the structure,
        modifications to the cutoff radius, or both.
        """
        if self.neighbor is not None:
            self.neighbor.update(self, r_cutoff)
        elif r_cutoff is not None:
            self.neighbor = Neighbor.from_structure(self, r_cutoff)
        else:
            logger.error(
                "No cutoff radius was found",
                exception=ValueError,
            )

    @classmethod
    def _get_atom_attributes(cls) -> Tuple[str, ...]:
        return cls._get_jit_dynamic_attributes()

    def __hash__(self) -> int:
        """Use parent class's hash method because of JIT."""
        return super().__hash__()

    @property
    def natoms(self) -> int:
        """Return number of atoms in the structure"""
        return self.positions.shape[0]

    @property
    def dtype(self) -> Dtype:
        """Return data type of the arrays in the structure (e.g., float64)."""
        return self.positions.dtype

    @property
    def r_cutoff(self) -> Optional[float]:
        """Return cutoff radius for neighboring atoms."""
        if self.neighbor is not None:
            return self.neighbor.r_cutoff

    @property
    def lattice(self) -> Optional[Array]:
        """Cell 3x3 matrix."""
        if self.box is not None:
            return self.box.lattice

    def get_unique_elements(self) -> Tuple[Element, ...]:
        return tuple(sorted(set(self.get_elements())))

    def get_elements(self) -> Tuple[Element, ...]:
        """Get array of elements."""
        to_element = self.element_map.atom_type_to_element
        atom_types_host = jax.device_get(self.atom_types)
        return tuple(str(to_element[int(at)]) for at in atom_types_host)

    def get_masses(self) -> Array:
        """Get array of atomic masses."""
        to_element = self.element_map.atom_type_to_element
        elements = (to_element[int(at)] for at in self.atom_types)
        return jnp.array(
            tuple(
                ElementMap.element_to_atomic_mass(element)
                for element in elements
            )
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}"
            f"(natoms={self.natoms}, "
            f"elements={self.get_unique_elements()}, "
            f"dtype={self.dtype})"
        )

    def select(self, element: Element) -> Array:
        """
        Retrieve the indices of all atoms that correspond to the given element.

        :param element: element name (e.g. `H` for hydrogen)
        :return: atom indices
        """
        return jnp.nonzero(
            self.atom_types == self.element_map.element_to_atom_type[element]
        )[0]

    @jax.jit
    def calculate_distances(
        self,
        atom_indices: Optional[Array] = None,
        neighbor_indices: Optional[Array] = None,
    ) -> Tuple[Array, Array]:
        """
        Calculate distances between specific atoms (given by atom indices)
        and the neighboring atoms in the structure.
        This method optionally also returns the corresponding position differences.

        If atom indices are not specified, all atoms in the structure will be taken into account.
        Similarly, if neighbor indices are not provided, all neighboring atoms will be considered.

        :param atom_indices: array of atom indices (zero-based index)
        :type atom_indices: Optional[Array], optional
        :param neighbor_indices: indices of neighbor atoms, defaults to None
        :type neighbor_indices: Optional[Array], optional
        :type neighbor_indices: bool, optional
        :param return_position_differences: whether returning position differences, defaults to False
        :type return_position_differences: bool, optional
        :return:  distances between atoms
        :rtype: Tuple[Array, ...]
        """
        if atom_indices is not None:
            atom_positions = self.positions[
                jnp.asarray([atom_indices])
            ].reshape(-1, 3)
        else:
            atom_positions = self.positions

        if neighbor_indices is not None:
            neighbor_positions = self.positions[
                jnp.atleast_1d(neighbor_indices)
            ]
        else:
            neighbor_positions = self.positions

        distances, position_differences = _calculate_distances(
            atom_positions,
            neighbor_positions,
            self.lattice,
        )
        return jnp.squeeze(distances), jnp.squeeze(position_differences)

    def to_dict(self) -> Dict[str, np.ndarray]:
        """
        The atomic attributes are represented as a dictionary of NumPy arrays.
        This format can be employed, for instance, when saving the structure data into a file.

        :return: dictionary of atom attributes.
        :rtype: Dict[str, np.ndarray]
        """
        data = dict()
        for atom_attr in self._get_atom_attributes():
            array: Array = getattr(self, atom_attr)
            data[atom_attr] = np.asarray(array)
        return data

    def to_ase(self) -> AseAtoms:
        """
        Represent the structure as ASE atoms.

        The returned object can be utilized with the `ASE`_ package
        for visualization or modification of the structure.

        :return: `ASE`_ representation of the structure

        .. _ASE: https://wiki.fysik.dtu.dk/ase/index.html
        """
        logger.info("Creating an ASE representation of the structure")
        to_element = self.element_map.atom_type_to_element
        return AseAtoms(
            symbols=[to_element[int(at)] for at in self.atom_types],
            positions=[
                units.TO_ANGSTROM * np.asarray(pos) for pos in self.positions
            ],
            cell=units.TO_ANGSTROM * np.asarray(self.box.lattice)
            if self.box is not None
            else None,
            pbc=True if self.box else False,
            charges=[np.asarray(ch) for ch in self.charges],
        )

    def _get_energy_offset(self, atom_energy: Dict[Element, float]) -> Array:
        energy_offset: Array = jnp.empty_like(self.energies)
        for element in self.get_unique_elements():
            energy_offset = energy_offset.at[self.select(element)].set(
                atom_energy[element]
            )
        return energy_offset

    def remove_energy_offset(self, atom_energy: Dict[Element, float]) -> None:
        """
        Remove the input reference energies from individual atoms and the total energy.

        :param atom_energy: atom reference energy
        :type atom_energy: Dict[Element, float]]
        """
        energy_offset: Array = self._get_energy_offset(atom_energy)
        self.energies -= energy_offset
        self.total_energy -= energy_offset.sum()

    def add_energy_offset(self, atom_energy: Dict[Element, float]) -> None:
        """
        Add the input reference energies to individual atoms and the total energy.

        :param atom_energy: atom reference energy
        :type atom_energy: Dict[Element, float]]
        """
        energy_offset = self._get_energy_offset(atom_energy)
        self.energies += energy_offset
        self.total_energy += energy_offset.sum()

    def get_center_of_mass_position(self) -> Array:
        """Get center of mass position."""
        return _calculate_center_of_mass(self.positions, self.get_masses())

    def get_per_element_inputs(self) -> Dict[Element, Inputs]:
        """Get required info per element for training and evaluating a potential."""

        def extract_inputs() -> Iterator[Tuple[Element, Inputs]]:
            for element in self.get_unique_elements():
                atom_indices: Array = self.select(element)
                yield element, Inputs(
                    self.positions[atom_indices],
                    self.positions,
                    self.atom_types,
                    self.lattice,
                    tree_util.tree_map(
                        lambda x: jnp.asarray(x),
                        self.element_map.element_to_atom_type,
                    ),
                )

        return {element: input for element, input in extract_inputs()}

    def get_per_element_positions(self) -> Dict[Element, Array]:
        """Get position of atoms per element."""

        def extract_positions() -> Iterator[Tuple[Element, Array]]:
            for element in self.get_unique_elements():
                atom_indices = self.select(element)
                yield element, self.positions[atom_indices]

        return {element: position for element, position in extract_positions()}

    def get_per_element_forces(self) -> Dict[Element, Array]:
        """Get force components per element."""

        def extract_forces() -> Iterator[Tuple[Element, Array]]:
            for element in self.get_unique_elements():
                atom_indices = self.select(element)
                yield element, self.forces[atom_indices]

        return {element: force for element, force in extract_forces()}


class Inputs(NamedTuple):
    """Represent array data of Structure for computing energy and forces."""

    atom_positions: Array
    positions: Array
    atom_types: Array
    lattice: Array
    emap: Dict[Element, Array]


register_jax_pytree_node(Structure)



import random
from collections import defaultdict
from typing import Callable, Dict

import jax
import jax.numpy as jnp
import numpy as np
from frozendict import frozendict
from jax import flatten_util
from pantea.atoms.structure import Structure
from pantea.datasets.dataset import DatasetInterface
from pantea.logger import logger
from pantea.potentials._energy import _compute_energy
from pantea.potentials._force import _compute_force
from pantea.potentials.atomic_potential import AtomicPotential
from pantea.potentials.nnp.nnp import (
    NeuralNetworkPotentialInterface as PotentialInterface,
)
from pantea.potentials.nnp.settings import (
    NeuralNetworkPotentialSettings as PotentialSettings,
)
from pantea.types import Array, Element, default_dtype
from tqdm import tqdm


def _tree_flatten(pytree: Dict) -> Array:
    return flatten_util.ravel_pytree(pytree)[0].reshape(-1, 1)  # type: ignore


class KalmanFilterUpdater:
    """
    A potential trainer which uses Kalman filter to update model weights (see this_).

    .. _this: https://pubs.acs.org/doi/10.1021/acs.jctc.8b01092
    """

    # https://github.com/CompPhysVienna/n2p2/blob/master/src/libnnptrain/KalmanFilter.cpp

    def __init__(self, potential: PotentialInterface) -> None:
        self.potential: PotentialInterface = potential
        self._init_parameters()
        self._init_matrices()

    def _init_parameters(self) -> None:
        """Set required parameters from the potential settings."""
        settings: PotentialSettings = self.potential.settings
        self.kalman_type: int = settings.kalman_type
        self.beta: float = settings.force_weight
        self.force_fraction: float = settings.short_force_fraction
        # Standard
        self.epsilon: float = settings.kalman_epsilon
        self.q: float = settings.kalman_q0
        self.q_tau: float = settings.kalman_qtau
        self.q_min: float = settings.kalman_qmin
        self.eta: float = settings.kalman_eta
        self.eta_tau: float = settings.kalman_eta
        self.eta_max: float = settings.kalman_etamax
        # Fading memory
        if self.kalman_type == 1:
            self.lambda0: float = settings.kalman_lambda_short
            self.nu: float = settings.kalman_neu_short
            self.gamma: float = 1.0

    def _init_matrices(self) -> None:
        """Initialize required matrices."""
        model_params: Dict[str, frozendict] = self.potential.model_params

        # Initialize state vector
        W, tree_unflatten = flatten_util.ravel_pytree(model_params)  # type: ignore
        self.W: Array = W.reshape(-1, 1)
        self._unflatten_state_vector: Callable = tree_unflatten
        self.num_states: int = self.W.shape[0]
        # Error covariance matrix
        self.P: Array = (1.0 / self.epsilon) * jnp.identity(
            self.num_states, dtype=default_dtype.FLOATX
        )

    def fit(
        self,
        dataset: DatasetInterface,
    ) -> defaultdict:
        """
        Fit potential weights.

        :param dataset: training dataset
        :type dataset: StructureDataset
        """
        settings: PotentialSettings = self.potential.settings

        atomic_potential: Dict[
            Element, AtomicPotential
        ] = self.potential.atomic_potential

        model_params: Dict[Element, frozendict] = self.potential.model_params

        # ----------------------

        def compute_energy_error(
            model_params: Dict[Element, frozendict], structure: Structure
        ) -> Array:
            E_ref: Array = structure.total_energy
            E_pot: Array = _compute_energy(
                frozendict(atomic_potential),
                structure.get_per_element_positions(),
                model_params,
                structure.get_per_element_inputs(),
            )
            return (E_ref - E_pot)[0] / structure.natoms

        def compute_force_error(
            state_vector: Array, structure: Structure
        ) -> Array:
            model_params: Dict = self._unflatten_state_vector(state_vector)
            F_ref: Array = _tree_flatten(structure.get_per_element_forces())
            F_pot: Array = _tree_flatten(
                _compute_force(
                    frozendict(atomic_potential),
                    structure.get_per_element_positions(),
                    model_params,
                    structure.get_per_element_inputs(),
                )
            )
            return (F_ref - F_pot)[..., 0]

        grad_energy_error: Callable = jax.grad(compute_energy_error)
        jacob_force_error: Callable = jax.jacrev(compute_force_error)

        def compute_energy_error_gradient(
            model_params: Dict[Element, frozendict], structure: Structure
        ) -> Array:
            return _tree_flatten(
                grad_energy_error(model_params, structure),
            )

        def compute_force_error_jacobian(
            state_vector: Array, structure: Structure
        ) -> Array:
            return jacob_force_error(
                state_vector[..., 0], structure
            ).transpose()

        # ----------------------

        indices: list[int] = [i for i in range(len(dataset))]

        history = defaultdict(list)
        for epoch in range(settings.epochs):
            print(f"Epoch: {epoch + 1} of {settings.epochs}")
            random.shuffle(indices)

            loss_energy_per_epoch: Array = jnp.asarray(0.0)
            loss_force_per_epoch: Array = jnp.asarray(0.0)
            num_energy_updates_per_epoch: int = 0
            num_force_updates_per_epoch: int = 0

            for index in tqdm(indices):
                structure: Structure = dataset[index]

                # Error and jacobian matrices
                if np.random.rand() < self.force_fraction:
                    Xi = self.beta * compute_force_error(
                        self.W,
                        structure,
                    ).reshape(-1, 1)
                    H = -self.beta * compute_force_error_jacobian(
                        self.W,
                        structure,
                    )
                    loss_force_per_epoch += jnp.matmul(Xi.transpose(), Xi)[
                        0, 0
                    ]
                    num_force_updates_per_epoch += 1
                else:
                    Xi = compute_energy_error(model_params, structure).reshape(
                        -1, 1
                    )
                    H = -compute_energy_error_gradient(model_params, structure)
                    loss_energy_per_epoch += jnp.matmul(Xi.transpose(), Xi)[
                        0, 0
                    ]
                    num_energy_updates_per_epoch += 1

                num_observations: int = Xi.shape[0]

                # A temporary matrix
                X = self.P @ H

                # Scaling matrix
                A = H.transpose() @ X

                # Update learning rate
                if (self.kalman_type == 0) and (self.eta < self.eta_max):
                    self.eta *= np.exp(self.eta_tau)

                # Measurement noise
                if self.kalman_type == 0:
                    A += (1.0 / self.eta) * jnp.identity(
                        num_observations, dtype=default_dtype.FLOATX
                    )
                elif self.kalman_type == 1:
                    A += self.lambda0 * jnp.identity(
                        num_observations, dtype=default_dtype.FLOATX
                    )

                # Kalman gain matrix
                K = X @ jnp.linalg.inv(A)

                # Update error covariance matrix
                self.P -= K @ X.transpose()

                # Forgetting factor
                if self.kalman_type == 1:
                    self.P *= 1.0 / self.lambda0

                # Process noise.
                self.P += self.q * jnp.identity(
                    self.num_states, dtype=default_dtype.FLOATX
                )

                # Update state vector
                self.W += K @ Xi

                # Anneal process noise
                if self.q > self.q_min:
                    self.q *= np.exp(-self.q_tau)

                # Update forgetting factor
                if self.kalman_type == 1:
                    self.lambda0 = self.nu * self.lambda0 + 1.0 - self.nu
                    self.gamma = 1.0 / (1.0 + self.lambda0 / self.gamma)

                # Get params from state vector
                model_params = self._unflatten_state_vector(self.W)

            # Update model params
            logger.debug(f"Updating potential weights after epoch {epoch + 1}")
            self.potential.model_params = model_params

            loss_energy_per_epoch /= num_energy_updates_per_epoch
            loss_force_per_epoch /= num_force_updates_per_epoch
            loss_per_epoch = loss_energy_per_epoch + loss_force_per_epoch
            num_updates_per_epoch = (
                num_energy_updates_per_epoch + num_force_updates_per_epoch
            )

            print(
                f"training loss:{float(loss_per_epoch): 0.7f}"
                f", loss_energy:{float(loss_energy_per_epoch): 0.7f}"
                f", loss_force:{float(loss_force_per_epoch): 0.7f}"
            )
            history["epoch"].append(epoch + 1)
            history["loss"].append(loss_per_epoch)

            logger.info(
                f"energy_update_ratio: {num_energy_updates_per_epoch/num_updates_per_epoch:.3f}"
                f", force_update_ratio: {num_force_updates_per_epoch/num_updates_per_epoch:.3f}"
            )

        return history



from __future__ import annotations

import copy
import functools
import time
from collections import defaultdict
from typing import Any, Callable, Mapping

import pandas as pd
from pantea.logger import logger

_default_profiler = None


class Profiler:
    """
    An implementation of a basic profiler for debugging.

    How to use it:
        1) add profile() as a decorator before each targeted method
        2) use context manager to collect stats from the region of interest
    """

    def __init__(
        self, name: str = "Profiler", sort_by: str = "cumtime"
    ) -> None:
        logger.debug(
            f"An instance of {self.__class__.__name__} object has been initialized"
        )
        self.name: str = name
        self.sort_by: str = sort_by
        self.active: bool = False
        self._walltimer: float = 0.0
        self.stats: Mapping[str, Any] = defaultdict(
            lambda: {
                "number_of_calls": 0,
                "cumulative_time": 0.0,
                "star_walltime": None,
                "end_walltime": 0.0,
            }
        )

    def __enter__(self) -> Profiler:
        """Start profiling."""
        logger.info(f"{self.name} started")
        self.active = True
        global _default_profiler
        _default_profiler = self
        self._walltimer = time.perf_counter()
        return self

    def __exit__(self, type, value, traceback) -> None:
        """Stop profiling."""
        self.active = False
        global _default_profiler
        self.stats = copy.deepcopy(_default_profiler.stats)
        _default_profiler = None
        logger.info(f"{self.name} finished")
        logger.info(self)

    @staticmethod
    def profile(func) -> Callable:
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs) -> Any:
            global _default_profiler
            if _default_profiler is not None and _default_profiler.active:
                start_time: float = time.perf_counter()
                retval = func(self, *args, **kwargs)
                end_time: float = time.perf_counter()
                func_info: str = f"{self.__class__.__name__}.{func.__name__}"
                _default_profiler.stats[func_info]["number_of_calls"] += 1
                _default_profiler.stats[func_info]["cumulative_time"] += (
                    end_time - start_time
                )
                if (
                    _default_profiler.stats[func_info]["start_walltime"]
                    is None
                ):
                    _default_profiler.stats[func_info]["start_walltime"] = (
                        time.perf_counter() - _default_profiler._walltimer
                    )
                _default_profiler.stats[func_info]["end_walltime"] = (
                    time.perf_counter() - _default_profiler._walltimer
                )
            else:
                retval = func(self, *args, **kwargs)
            return retval

        return wrapper

    def get_dataframe(self) -> pd.DataFrame:
        """Return a dataframe representation of profiler statistics."""
        if len(self.stats) == 0:
            pd.DataFrame({})

        df = defaultdict(list)
        for func, stats in self.stats.items():
            df["method"].append(func)
            for metric, value in stats.items():
                df[metric].append(value)
        return pd.DataFrame(df).sort_values(self.sort_by, ascending=False)

    def __repr__(self) -> str:
        df_: pd.DataFrame = self.get_dataframe()
        if df_ is not None:
            return f"Profiler statistics:\n{df_.to_string()}"


class Timer:
    """A class to measure elapsed time."""

    def __init__(self, name: str = "Timer") -> None:
        self.name: str = name
        self.elapsed_time: float = 0.0
        self._start: float

    def __enter__(self) -> Timer:
        self._start = time.perf_counter()
        return self

    def __exit__(self, type, value, traceback) -> None:
        self.elapsed_time = time.perf_counter() - self._start
        print(self)

    def __repr__(self) -> str:
        return f"{self.name} (elapsed time {self.elapsed_time:.8f} seconds)"


def timer(func: Callable):
    """A decorator to measure elapsed time when calling a function."""
    print(func.__name__)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with Timer(f"Timer {func.__name__} function"):
            retval = func(*args, **kwargs)
        return retval

    return wrapper



from typing import Any, Protocol

from pantea.types import Array


class ModelInterface(Protocol):
    def __call__(self, inputs: Array) -> Array:
        """Compute energy."""
        ...

    def save(self, *arg: Any, **kwargs: Any) -> None:
        """Save model weights."""
        ...

    def load(self, *arg: Any, **kwargs: Any) -> Any:
        """Load model weights."""
        ...



from functools import partial
from typing import Callable, Dict, Protocol, Tuple

import jax
import jax.numpy as jnp
from jax import jit, lax, vmap

from pantea.atoms._structure import _calculate_distances_per_atom
from pantea.atoms.neighbor import _calculate_masks_per_atom
from pantea.descriptors.acsf.angular import AngularSymmetryFunction
from pantea.descriptors.acsf.radial import RadialSymmetryFunction
from pantea.descriptors.acsf.symmetry import EnvironmentElements
from pantea.types import Array, Element


@jit
def _calculate_radial_acsf_per_atom(
    radial: Dict[EnvironmentElements, RadialSymmetryFunction],
    atype: Array,
    dist_i: Array,
    emap: Dict[Element, Array],
) -> Array:
    elements: EnvironmentElements = [k for k in radial.keys()][0]

    mask_cutoff_i = _calculate_masks_per_atom(
        dist_i, jnp.asarray(radial[elements].r_cutoff)
    )
    mask_cutoff_and_atype_ij = mask_cutoff_i & (
        atype == emap[elements.neighbor_j]
    )

    return jnp.sum(
        radial[elements](dist_i),
        where=mask_cutoff_and_atype_ij,
        axis=0,
    )


@jit
def _calculate_angular_acsf_per_atom(
    angular: Dict[EnvironmentElements, AngularSymmetryFunction],
    atype: Array,
    diff_i: Array,
    dist_i: Array,
    lattice: Array,
    emap: Dict[Element, Array],
) -> Array:
    elements: EnvironmentElements = [k for k in angular.keys()][0]

    # cutoff-radius mask
    mask_cutoff_i = _calculate_masks_per_atom(
        dist_i, jnp.asarray(angular[elements].r_cutoff)
    )
    # mask for neighboring element j
    at_j = emap[elements.neighbor_j]
    mask_cutoff_and_atype_ij = mask_cutoff_i & (atype == at_j)
    # mask for neighboring element k
    at_k = emap[elements.neighbor_k]  # type: ignore
    mask_cutoff_and_atype_ik = mask_cutoff_i & (atype == at_k)

    total, _ = lax.scan(
        partial(
            _inner_loop_over_angular_acsf_terms,
            mask_ik=mask_cutoff_and_atype_ik,
            diff_i=diff_i,
            dist_i=dist_i,
            lattice=lattice,
            kernel=angular[elements],
        ),
        jnp.array(0.0),
        (diff_i, dist_i, mask_cutoff_and_atype_ij),
    )

    # correct double-counting
    total = lax.cond(
        at_j == at_k,
        lambda x: x * 0.5,
        lambda x: x,
        total,
    )
    return total


# Called by jax.lax.scan (no need for @jax.jit)
def _inner_loop_over_angular_acsf_terms(
    total: Array,
    inputs: Tuple[Array, ...],
    diff_i: Array,
    dist_i: Array,
    mask_ik: Array,
    lattice: Array,
    kernel: Callable,
) -> Tuple[Array, Array]:
    # Scan occurs along the leading axis
    Rij, rij, mask_ij = inputs

    # rik = jnp.where(
    #     mask_cutoff_and_atype_ik,
    #     dist_i,
    #     0.0,
    # ) # same for Rjk if using diff_i

    # fix nan issue in gradient
    # see https://github.com/google/jax/issues/1052#issuecomment-514083352
    operand = rij * dist_i
    is_zero = operand == 0.0
    true_op = jnp.where(is_zero, 1.0, operand)
    cost = jnp.where(
        mask_ik,
        jnp.inner(Rij, diff_i) / true_op,
        1.0,
    )
    cost = jnp.where(is_zero, 0.0, cost)  # type: ignore

    rjk = jnp.where(  # diff_jk = diff_ji - diff_ik
        mask_ik,
        _calculate_distances_per_atom(Rij, diff_i, lattice)[0],
        0.0,
    )  # second tuple output for Rjk

    value = jnp.where(
        mask_ij,
        jnp.sum(
            kernel(rij, dist_i, rjk, cost),
            where=mask_ik & (rjk > 0.0),  # exclude k=j # type:ignore
            axis=0,
        ),
        0.0,
    )

    return total + value, value


class AcsfInterface(Protocol):
    num_symmetry_functions: int
    num_radial_symmetry_functions: int
    num_angular_symmetry_functions: int
    radial_symmetry_functions: Tuple
    angular_symmetry_functions: Tuple


@jit
def _calculate_descriptor_per_atom(
    acsf: AcsfInterface,
    single_atom_position: Array,
    neighbor_positions: Array,
    atype: Array,
    lattice: Array,
    emap: Dict,
) -> Array:
    """
    Compute descriptor values per atom in the structure (via atom id).
    """
    dtype = single_atom_position.dtype
    result: Array = jnp.empty(acsf.num_symmetry_functions, dtype=dtype)

    dist_i, diff_i = _calculate_distances_per_atom(
        single_atom_position, neighbor_positions, lattice
    )

    # Loop over the radial terms
    for index, (elements, radial) in enumerate(acsf.radial_symmetry_functions):
        result = result.at[index].set(
            _calculate_radial_acsf_per_atom(
                {elements: radial}, atype, dist_i, emap
            )
        )

    # Loop over the angular terms
    for index, (elements, angular) in enumerate(
        acsf.angular_symmetry_functions,
        start=acsf.num_radial_symmetry_functions,
    ):
        result = result.at[index].set(
            _calculate_angular_acsf_per_atom(
                {elements: angular}, atype, diff_i, dist_i, lattice, emap
            )
        )

    return result


_calculate_descriptor = jit(
    vmap(
        _calculate_descriptor_per_atom,
        in_axes=(None, 0, None, None, None, None),
    ),
    static_argnums=(0,),
)


_calculate_grad_descriptor_per_atom = jit(
    jax.jacfwd(
        _calculate_descriptor_per_atom,
        argnums=1,
    ),
    static_argnums=(0,),
)



import math
from typing import Generator

from pantea.types import Array


def create_batch(
    array: Array,
    batch_size: int,
) -> Generator[Array, None, None]:
    """
    Create baches of the input array.

    :param array: input array
    :type array: Array
    :param batch_size: desired batch size
    :type batch_size: int
    :yield: a batch of input array
    :rtype: Generator[Array, None, None]
    """
    n_batches = int(math.ceil(len(array) / batch_size))  # type: ignore
    for i in range(n_batches):
        yield array[i * batch_size : (i + 1) * batch_size, ...]



from dataclasses import dataclass

from frozendict import frozendict

from pantea.atoms.structure import Structure
from pantea.descriptors.descriptor import DescriptorInterface
from pantea.descriptors.scaler import Scaler
from pantea.models.nn.network import NeuralNetworkModel
from pantea.potentials._energy import _compute_atomic_energy
from pantea.types import Array, Element


@dataclass(frozen=True)
class AtomicPotential:
    """
    Atomic potential.

    It chains all the required transformers (descriptor, scaler, model, etc.)
    to calculate per-atom energy.
    """

    descriptor: DescriptorInterface
    scaler: Scaler
    model: NeuralNetworkModel

    def apply(
        self,
        params: frozendict,
        structure: Structure,
    ) -> Array:
        """
        Calculate model output energy.
        It must be noted model output for each element has no physical meaning.
        It's only the total energy, sum of the atomic energies over all atom in the structure,
        which has actually physical meaning.

        :param params: model parameters
        :param structure: input structure
        :return: model energy output
        """
        element: Element = self.descriptor.central_element  # type: ignore
        aid: Array = structure.select(element)

        return _compute_atomic_energy(
            self,
            structure.positions[aid],
            params,
            structure.get_per_element_inputs()[element],
        )

    @property
    def model_input_size(self) -> int:
        """Return size of the model input."""
        return self.descriptor.num_descriptors

    def __repr__(self) -> str:
        out: str = f"{self.__class__.__name__}("
        for component in self.__annotations__:
            out += f"\n  {component}={getattr(self, component)},"
        out += "\n)"
        return out



from dataclasses import dataclass

import jax.numpy as jnp
from jax._src.numpy.lax_numpy import _ScalarMeta
from jaxlib.xla_extension import ArrayImpl

Array = ArrayImpl
Dtype = _ScalarMeta
Element = str
Scalar = ArrayImpl


@dataclass
class DataType:
    """
    A configuration for array data type.

    It is globally used as default dtype for arrays, indices, etc.
    User can modify any default dtype, for example, via setting the global
    floating point precision (`FLOATX`) to a single (float32)
    or double (float64).
    """

    FLOATX: Dtype = jnp.float32
    INT: Dtype = jnp.int32
    UINT: Dtype = jnp.uint32
    INDEX: Dtype = jnp.int32


default_dtype = DataType()



from typing import Protocol

from pantea.atoms.structure import Structure


class DatasetInterface(Protocol):
    """
    A data container for atom data structure.

    Features:

    * it must access data item in a lazy mode.
    * it should be able to cache data via a `persist` input flag.
    """

    def __len__(self) -> int:
        ...

    def __getitem__(self, index: int) -> Structure:
        ...



from __future__ import annotations

from typing import Any, Dict, Tuple

from jax import tree_util
from pantea.logger import logger


class BaseJaxPytreeDataClass:
    """
    Here we specify exactly which components of a `dataclass` should be treated as
    static and which should be treated as dynamic (array) attributes.

    The `hash` method then it would only based on the static attributes and
    used for JIT re-compilation if they change.

    .. warning::

        It must be noted that we have to define first dynamic values (as tuple)
        and then static values (as dict) in child class constructor (same for dataclasses).

        Registering the class to jax pytree node is also required.

        Must be used only for dataclasses.

    See https://jax.readthedocs.io/en/latest/faq.html#how-to-use-jit-with-methods
    """

    def _tree_flatten(self) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        children: Tuple[Any, ...] = tuple(
            getattr(self, attr) for attr in self._get_jit_dynamic_attributes()
        )
        aux_data: Dict[str, Any] = {
            attr: getattr(self, attr)
            for attr in self._get_jit_static_attributes()
        }
        return (children, aux_data)

    @classmethod
    def _tree_unflatten(
        cls, aux_data: Dict[str, Any], children: Tuple[Any, ...]
    ) -> BaseJaxPytreeDataClass:
        return cls(*children, **aux_data)  # type: ignore

    def __hash__(self) -> int:
        """
        Define hash based on the `JAX JIT compilation` static attributes.
        This is used for detecting changes on the static arguments
        of the jit-compiled method.
        """
        aux_data: Dict[str, Any] = self._tree_flatten()[1]
        return hash(tuple(frozenset(sorted(aux_data.items()))))

    @classmethod
    def _get_jit_dynamic_attributes(cls) -> Tuple[str, ...]:
        """Get JAX JIT compilation dynamic attribute names (i.e. jax.ndarray)."""
        return tuple(
            attr
            for attr, dtype in cls.__annotations__.items()
            if "Array" in str(dtype)
        )

    @classmethod
    def _get_jit_static_attributes(cls) -> Tuple[str, ...]:
        """Get JAX JIT compilation static attribute names."""
        dynamic_attributes = cls._get_jit_dynamic_attributes()
        return tuple(
            attr
            for attr in cls.__annotations__.keys()
            if attr not in dynamic_attributes
        )

    @classmethod
    def _assert_jit_attributes(
        cls,
        available: Tuple[str, ...],
        expected: Tuple[str, ...],
        tag: str = "",
    ) -> None:
        """Assert to ensure jit (static or dynamics) attributes are correctly identified."""
        if sorted(available) != sorted(expected):  # type: ignore
            logger.error(
                f"JIT {tag} attributes: expected {expected} but got {available}",
                exception=AssertionError,
            )

    @classmethod
    def _assert_jit_static_attributes(
        cls, expected: Tuple[str, ...] = tuple()
    ) -> None:
        cls._assert_jit_attributes(
            cls._get_jit_static_attributes(), expected, tag="static"
        )

    @classmethod
    def _assert_jit_dynamic_attributes(
        cls, expected: Tuple[str, ...] = tuple()
    ) -> None:
        cls._assert_jit_attributes(
            cls._get_jit_dynamic_attributes(), expected, tag="dynamic"
        )


def register_jax_pytree_node(cls) -> None:
    """Register the input class as internal JAX pytree node."""
    tree_util.register_pytree_node(
        cls, cls._tree_flatten, cls._tree_unflatten  # type: ignore
    )



# This file includes the package version info
from typing import Tuple


def _version_as_tuple(version_str) -> Tuple[int, ...]:
    return tuple(int(i) for i in version_str.split(".") if i.isdigit())


__version__ = "0.8.2"

# __version_info__: tuple[int, ...] = _version_as_tuple(__version__)



import os
from typing import Tuple

os.environ["JAX_ENABLE_X64"] = "1"
os.environ["JAX_PLATFORM_NAME"] = "cpu"

import jax.numpy as jnp
import pytest
from jax import random
from pantea.descriptors.scaler import Scaler
from pantea.types import Array, default_dtype
from pantea.utils.batch import create_batch

key = random.PRNGKey(2023)
key, subkey = random.split(key, 2)


class TestStructure:
    data_1: Array = random.uniform(key, shape=(50, 3))
    data_2: Array = random.normal(subkey, shape=(100, 10), dtype=jnp.float64)
    data_3: Array = random.uniform(key, shape=(30, 5))

    @pytest.mark.parametrize(
        "data, expected",
        [
            (
                data_1,
                (3, 50, default_dtype.FLOATX),
            ),
            (
                data_2,
                (10, 100, jnp.float64),
            ),
            (
                data_3,
                (5, 30, default_dtype.FLOATX),
            ),
        ],
    )
    def test_general_attributes(
        self,
        data: Array,
        expected: Tuple,
    ) -> None:
        scaler = Scaler()
        scaler.fit(data)
        assert scaler.dimension == expected[0]
        assert scaler.nsamples == expected[1]

    def fit_scaler(self, scaler: Scaler, data: Array, batch_size: int) -> None:
        for batch in create_batch(data, batch_size=batch_size):
            scaler.fit(batch)

    def compare(self, scaler: Scaler, data: Array) -> None:
        for name in ("mean", "min", "max"):
            assert jnp.allclose(
                getattr(data, name)(axis=0), getattr(scaler, name)
            )
        assert jnp.allclose(data.std(axis=0), scaler.sigma)

    @pytest.mark.parametrize(
        "data",
        [
            data_1,
            data_2,
            data_3,
        ],
    )
    def test_array_attributes(
        self,
        data: Array,
    ) -> None:
        for scale_type, batch_size in zip(
            ("scale_center", "scale", "center"),
            (7, 10, 1),
        ):
            scaler = Scaler(scale_type=scale_type)
            self.fit_scaler(scaler, data, batch_size=batch_size)
            self.compare(scaler, data)

    @pytest.mark.parametrize(
        "data",
        [
            data_1,
            data_2,
            data_3,
        ],
    )
    def test_scaler_warnings(
        self,
        data: Array,
    ) -> None:
        scaler = Scaler()
        scaler.fit(data)

        assert scaler.max_number_of_warnings is None
        assert scaler.number_of_warnings == 0
        scaler.set_max_number_of_warnings(2)
        assert scaler.max_number_of_warnings == 2

        out_range_data: Array = jnp.ones_like(data) * 100
        scaler(out_range_data)
        assert scaler.number_of_warnings == 0
        print(out_range_data)
        scaler(out_range_data, warnings=True)
        assert scaler.number_of_warnings == 1



"""
Pantea is an optimized Python library on basis of Google JAX that enables development of machine
learning inter-atomic potentials for use in computational physics, chemistry, and material science.
These potentials are necessary for conducting large-scale molecular dynamics simulations of complex
materials with ab initio accuracy.
"""
from pantea._version import __version__ as version

__author__ = """Hossein Ghorbanfekr"""
__email__ = "hgh.comphys@gmail.com"
__version__ = version



import jax.numpy as jnp
from flax import linen as nn
from frozendict import frozendict
from pantea.types import Array


def identity(x: Array) -> Array:
    return x


def tanh(x: Array) -> Array:
    return nn.tanh(x)


def logistic(x: Array) -> Array:
    return 1.0 / (1.0 + jnp.exp(-x))


def softplus(x: Array) -> Array:
    return nn.softplus(x)


def relu(x: Array) -> Array:
    return nn.relu(x)


def gaussian(x: Array) -> Array:
    return jnp.exp(-0.5 * x**2)


def cos(x: Array) -> Array:
    return jnp.cos(x)


def revlogistic(x: Array) -> Array:
    return 1.0 - 1.0 / (1.0 + jnp.exp(-x))


def exp(x: Array) -> Array:
    return jnp.exp(-x)


def harmonic(x: Array) -> Array:
    return x * x


# see here https://compphysvienna.github.io/n2p2/api/neural_network.html?highlight=activation%20function
_activation_function_map: frozendict = frozendict(
    {
        "identity": identity,
        "tanh": nn.tanh,
        "logistic": logistic,
        "softplus": nn.softplus,
        "relu": nn.relu,
        "gaussian": gaussian,
        "cos": cos,
        "exp": exp,
        "harmonic": harmonic,
    }
)



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass

import jax
import jax.numpy as jnp
from pantea.descriptors.acsf.cutoff import CutoffFunction
from pantea.descriptors.acsf.symmetry import BaseSymmetryFunction
from pantea.pytree import register_jax_pytree_node
from pantea.types import Array


class RadialSymmetryFunction(BaseSymmetryFunction, metaclass=ABCMeta):
    """A base class for `two body` (radial) symmetry functions."""

    def __hash__(self) -> int:
        """Enforce to use the parent class's hash method (JIT)."""
        return super().__hash__()

    @abstractmethod
    def __call__(self, rij: Array) -> Array:
        ...


@dataclass
class G1(RadialSymmetryFunction):
    """Plain cutoff function as symmetry function."""

    cfn: CutoffFunction

    def __hash__(self) -> int:
        """Enforce to use the parent class's hash method (JIT)."""
        return super().__hash__()

    @jax.jit
    def __call__(self, rij: Array) -> Array:
        return self.cfn(rij)


@dataclass
class G2(RadialSymmetryFunction):
    """Radial exponential symmetry function."""

    cfn: CutoffFunction
    r_shift: float
    eta: float

    def __hash__(self) -> int:
        """Enforce to use the parent class's hash method (JIT)."""
        return super().__hash__()

    @jax.jit
    def __call__(self, rij: Array) -> Array:
        return jnp.exp(-self.eta * (rij - self.r_shift) ** 2) * self.cfn(rij)


register_jax_pytree_node(G1)
register_jax_pytree_node(G2)



from pantea.potentials.nnp.potential import NNP, NeuralNetworkPotential
from pantea.potentials.nnp.settings import NeuralNetworkPotentialSettings

__all__ = [
    "NeuralNetworkPotential",
    "NeuralNetworkPotentialSettings",
    "NNP",
]



from typing import Any, Dict, Optional

import jax.numpy as jnp
from pantea.logger import logger
from pantea.types import Array, Dtype, default_dtype


def set_as_attribute(
    obj: Any,
    items: Dict[str, Any],
    prefix: str = "",
    postfix: str = "",
) -> None:
    """
    An utility function to set a dictionary of items as the input object attributes.


    :param obj: instance
    :param items: dictionary of attributes
    :type items: Dict[str, Any]
    :param prefix: _description_, defaults to ""
    :type prefix: str, optional
    :param postfix: _description_, defaults to ""
    :type postfix: str, optional
    """
    logger.debug(
        f"Setting {len(items)} items as {obj.__class__.__name__} attributes:"
    )
    for name, item in items.items():
        attr_name: str = f"{prefix}{name}{postfix}"
        logger.debug(f"-> {obj.__class__.__name__}.{attr_name}")
        setattr(obj, attr_name, item)


def asarray(data: Any, dtype: Optional[Dtype] = None) -> Array:
    """
    An utility function to cast input data (scalar, array, etc) to Array type with predefined dtype.

    :param value: input data
    :type value: Any
    :param dtype: casted dtype. Default dtype will be used otherwise.
    :type dtype: Optional[Dtype], optional
    :return: casted input
    :rtype: Array
    """
    if dtype is None:
        dtype = default_dtype.FLOATX
    return jnp.asarray(data, dtype=dtype)



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, List, Mapping, NamedTuple, Union

from pantea.atoms.element import ElementMap
from pantea.config import _CFG
from pantea.logger import logger
from pantea.types import Element
from pantea.utils.tokenize import tokenize
from pydantic import Field, ValidationError, validator


class RadialSymFuncArgs(NamedTuple):
    """Symmetry function arguments."""

    central_element: str
    acsf_type: int
    neighbor_element_j: str
    eta: float
    r_cutoff: float
    r_shift: float


class AngularSymFuncArgs(NamedTuple):
    """Symmetry function arguments."""

    central_element: str
    acsf_type: int
    neighbor_element_j: str
    eta: float
    r_cutoff: float
    r_shift: float
    neighbor_element_k: str
    lambda0: float
    zeta: float


SymFuncArgs = Union[RadialSymFuncArgs, AngularSymFuncArgs]


cutoff_function_map: Mapping[str, str] = {
    "0": "hard",
    "1": "cos",
    "2": "tanhu",
    "3": "tanh",
    "4": "exp",
    "5": "poly1",
    "6": "poly2",
}

scaler_type_map: Mapping[str, str] = {
    "center_symmetry_functions": "center",
    "scale_symmetry_functions": "scale",
    "scale_center_symmetry_functions": "scale_center",
    "scale_center_symmetry_functions_sigma": "scale_center_sigma",
}

activation_function_map: Mapping[str, str] = {
    "l": "identity",
    "t": "tanh",
    "s": "logistic",
    "p": "softplus",
    "r": "relu",
    "g": "gaussian",
    "c": "cos",
    "e": "exp",
    "h": "harmonic",
}

updater_type_map: Mapping[str, str] = {
    "0": "gradient_descent",
    "1": "kalman_filter",
}

gradient_type_map: Mapping[str, str] = {
    "0": "fixed_step",
    "1": "adam",
}


class NeuralNetworkPotentialSettings(_CFG):
    """
    A configuration class for neural network potential parameters.

    It contains a collections of all potential setting keywords and their default values.
    """

    # sphinx-pydantic can be useful to better document pydantic classes
    # https://pypi.org/project/autodoc-pydantic/0.1.1/

    # General
    random_seed: int = Field(
        2023, description="determine initialize neural network weights."
    )

    number_of_elements: int
    elements: List[Element] = Field(default_factory=list)
    atom_energy: Dict[Element, float] = Field(default_factory=dict)
    scaler_save_naming_format: str = "scaling.{:03d}.data"
    model_save_naming_format: str = "weights.{:03d}.pkl"
    # Neural Network
    weights_min: float = -1.0
    weights_max: float = 1.0
    global_hidden_layers_short: int
    global_nodes_short: List[int]
    global_activation_short: List[str]
    # Trainer
    epochs: int = 1
    updater_type: str = "gradient_descent"
    gradient_type: str = "adam"
    main_error_metric: str = "RMSE"
    force_weight: float = 1.0
    short_force_fraction: float = 0.1
    short_energy_fraction: float = 1.0
    test_fraction: float = 0.1
    save_best_model: bool = True
    gradient_eta: float = 1.0e-5
    gradient_adam_eta: float = 1.0e-3
    gradient_adam_beta1: float = 0.9
    gradient_adam_beta2: float = 0.999
    gradient_adam_epsilon: float = 1.0e-8
    gradient_adam_weight_decay: float = 1.0e-4
    kalman_type: int = 0
    kalman_epsilon: float = 0.01
    kalman_q0: float = 0.01
    kalman_qtau: float = 2.302
    kalman_qmin: float = 1.0e-6
    kalman_eta: float = 0.01
    kalman_etatau: float = 2.302
    kalman_etamax: float = 1.0
    kalman_lambda_short: float = 0.96000
    kalman_neu_short: float = 0.99950
    # Symmetry Function
    cutoff_type: str = "tanh"
    scale_type: str = "center"
    scale_min_short: float = 0.0
    scale_max_short: float = 1.0
    symfunction_short: List[SymFuncArgs] = Field(default_factory=list)

    @classmethod
    def from_file(cls, filename: Path) -> NeuralNetworkPotentialSettings:
        """
        Read all potential settings from the input file.

        Parameters such as `elements`, `cutoff type`, `symmetry functions`, `neural network`, `training parameters`, etc
        (see `here <https://compphysvienna.github.io/n2p2/topics/keywords.html>`_).
        """

        logger.info(f"Reading potential settings: {str(filename)}")
        dict_: Dict = cls._read_from_file(filename)
        kwargs: Dict[str, Any] = dict()
        kwargs["atom_energy"] = dict()
        kwargs["symfunction_short"] = list()

        for line_keyword, tokens in dict_.items():
            keyword: str = "_".join(line_keyword.split("_")[1:])  # type: ignore
            # ------------- General -------------
            if keyword == "number_of_elements":  # this keyword can be ignored
                kwargs[keyword] = tokens[0]
            elif keyword == "elements":
                kwargs[keyword] = sorted(
                    set([t for t in tokens]), key=ElementMap.get_atomic_number
                )
            elif keyword == "atom_energy":
                kwargs[keyword].update({tokens[0]: tokens[1]})
            elif keyword == "random_seed":
                kwargs[keyword] = tokens[0]
            # ------------- Neural Network -------------
            elif keyword == "global_hidden_layers_short":
                kwargs[keyword] = tokens[0]
            elif keyword == "global_nodes_short":
                kwargs[keyword] = [t for t in tokens]
            elif keyword == "global_activation_short":
                kwargs[keyword] = [activation_function_map[t] for t in tokens]
            elif keyword == "weights_min":
                kwargs[keyword] = tokens[0]
            elif keyword == "weights_max":
                kwargs[keyword] = tokens[0]
            # ------------- Trainer -------------
            elif keyword == "main_error_metric":
                kwargs[keyword] = tokens[0]
            elif keyword == "epochs":
                kwargs[keyword] = tokens[0]
            elif keyword == "test_fraction":
                kwargs[keyword] = tokens[0]
            elif keyword == "updater_type":
                kwargs[keyword] = updater_type_map[tokens[0]]
            elif keyword == "gradient_type":
                kwargs[keyword] = gradient_type_map[tokens[0]]
            elif keyword == "gradient_eta":
                kwargs[keyword] = tokens[0]
            elif keyword == "gradient_adam_eta":
                kwargs[keyword] = tokens[0]
            elif keyword == "gradient_adam_beta1":
                kwargs[keyword] = tokens[0]
            elif keyword == "gradient_adam_beta2":
                kwargs[keyword] = tokens[0]
            elif keyword == "gradient_adam_epsilon":
                kwargs[keyword] = tokens[0]
            elif keyword == "gradient_adam_weight_decay":
                kwargs[keyword] = tokens[0]
            elif keyword == "force_weight":
                kwargs[keyword] = tokens[0]
            elif keyword == "short_force_fraction":
                kwargs[keyword] = tokens[0]
            elif keyword == "short_energy_fraction":
                kwargs[keyword] = tokens[0]
            elif keyword == "kalman_type":
                kwargs[keyword] = tokens[0]
            elif keyword == "kalman_epsilon":
                kwargs[keyword] = tokens[0]
            elif keyword == "kalman_q0":
                kwargs[keyword] = tokens[0]
            elif keyword == "kalman_qtau":
                kwargs[keyword] = tokens[0]
            elif keyword == "kalman_qmin":
                kwargs[keyword] = tokens[0]
            elif keyword == "kalman_eta":
                kwargs[keyword] = tokens[0]
            elif keyword == "kalman_etatau":
                kwargs[keyword] = tokens[0]
            elif keyword == "kalman_etamax":
                kwargs[keyword] = tokens[0]
            elif keyword == "kalman_lambda_short":
                kwargs[keyword] = tokens[0]
            elif keyword == "kalman_neu_short":
                kwargs[keyword] = tokens[0]
            # ------------- Symmetry Function -------------
            elif keyword == "cutoff_type":
                kwargs[keyword] = cutoff_function_map[tokens[0]]
            elif keyword == "center_symmetry_functions":
                kwargs["scale_type"] = scaler_type_map[keyword]
            elif keyword == "scale_symmetry_functions":
                kwargs["scale_type"] = scaler_type_map[keyword]
            elif keyword == "scale_center_symmetry_functions":
                kwargs["scale_type"] = scaler_type_map[keyword]
            elif keyword == "scale_center_symmetry_functions_sigma":
                kwargs["scale_type"] = scaler_type_map[keyword]
            elif keyword == "scale_min_short":
                kwargs[keyword] = tokens[0]
            elif keyword == "scale_max_short":
                kwargs[keyword] = tokens[0]
            elif keyword == "symfunction_short":
                acsf_type = int(tokens[1])
                args: SymFuncArgs
                if acsf_type < 3:  # radial
                    args = RadialSymFuncArgs(
                        central_element=tokens[0],
                        acsf_type=acsf_type,
                        neighbor_element_j=tokens[2],
                        eta=float(tokens[3]),
                        r_shift=float(tokens[4]),
                        r_cutoff=float(tokens[5]),
                    )
                else:  # angular
                    args = AngularSymFuncArgs(
                        central_element=tokens[0],
                        acsf_type=acsf_type,
                        neighbor_element_j=tokens[2],
                        neighbor_element_k=tokens[3],
                        eta=float(tokens[4]),
                        lambda0=float(tokens[5]),
                        zeta=float(tokens[6]),
                        r_cutoff=float(tokens[7]),
                        r_shift=float(tokens[8]) if len(tokens) == 9 else 0.0,
                    )
                kwargs[keyword].append(args)
        try:
            settings = cls(**kwargs)
        except ValidationError as e:
            logger.error(str(e), exception=ValueError)
        return settings  # type: ignore

    @classmethod
    def _read_from_file(cls, filename) -> Dict:
        """Read all keywords from the input setting file."""
        dict_ = dict()
        n_line: int = 0
        with open(str(filename), "r") as file:
            while True:
                line = file.readline()
                n_line += 1
                if not line:
                    break
                # Read keyword and values
                keyword, tokens = tokenize(line, comment="#")  # type: ignore
                if keyword is not None:
                    if keyword not in cls.__annotations__:
                        logger.debug(
                            f"Unknown keyword (line {n_line}):'{keyword}'"
                        )
                    else:
                        logger.debug(f"keyword:'{keyword}', tokens:{tokens}")
                        dict_[f"line{n_line:04d}_{keyword}"] = tokens
        return dict_

    @validator("elements")
    def number_of_elements_match(cls, v, values) -> Any:
        if (
            "number_of_elements" in values
            and len(v) != values["number_of_elements"]
        ):
            raise ValueError("number of elements is not consistent")
        return v

    @validator("test_fraction")
    def test_fraction_range(cls, v) -> Any:
        assert 0.0 <= v <= 1.0, "must be between [0, 1]"
        return v



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Optional, Protocol, Tuple

import jax
import jax.numpy as jnp

from pantea.atoms._structure import _calculate_distances
from pantea.logger import logger
from pantea.pytree import BaseJaxPytreeDataClass, register_jax_pytree_node
from pantea.types import Array


class Structure(Protocol):
    positions: Array
    lattice: Array


# @jax.jit
def _calculate_masks_per_atom(
    rij: Array,
    r_cutoff: Array,
) -> Array:
    """Return masks (boolean array) of a single atom inside a cutoff radius."""
    return (rij <= r_cutoff) & (rij != 0.0)


_vmap_calculate_masks: Callable = jax.vmap(
    _calculate_masks_per_atom,
    in_axes=(0, None),
)


# @jax.jit
def _calculate_masks(
    rij: Array,
    r_cutoff: Array,
) -> Array:
    """Calculate masks (boolean arrays) of multiple atoms inside a cutoff radius."""
    return _vmap_calculate_masks(rij, r_cutoff)


@jax.jit
def _calculate_masks_and_distances(
    structure: Structure,
    r_cutoff: Array,
) -> Tuple[Array, Array, Array]:
    """Calculate masks (boolean arrays) of multiple atoms inside a cutoff radius."""
    rij, Rij = _calculate_distances(
        atom_positions=structure.positions,
        neighbor_positions=structure.positions,
        lattice=structure.lattice,
    )
    masks = _calculate_masks(rij, r_cutoff)
    return masks, rij, Rij


@dataclass
class Neighbor(BaseJaxPytreeDataClass):
    """
    Create a neighbor list of atoms for structure.

    .. note::
        For MD simulations, re-neighboring the list is required every few steps.
        This is usually implemented together with defining a skin radius.
    """

    r_cutoff: float
    masks: Array
    rij: Array
    Rij: Array

    def __post_init__(self) -> None:
        """Post initialize the neighbor list."""
        # logger.debug(f"Initializing {self}")
        self._assert_jit_dynamic_attributes(expected=("masks", "rij", "Rij"))
        self._assert_jit_static_attributes(expected=("r_cutoff",))

    @classmethod
    def from_structure(cls, structure, r_cutoff: float) -> Neighbor:
        results = _calculate_masks_and_distances(
            structure, jnp.atleast_1d(r_cutoff)
        )
        return cls(r_cutoff, *results)

    # @jax.jit
    def update(
        self,
        structure: Structure,
        r_cutoff: Optional[float] = None,
    ) -> None:
        """
        Update neighboring atoms.

        This approach relies on cutoff masks, which is different from conventional
        methods used to update the neighbor list (such as defining neighbor indices).
        The rationale behind this approach is that JAX executes efficiently on
        vectorized variables, offering faster performance compared to simple Python loops.
        """
        logger.debug(f"Updating neighbor list ({r_cutoff=})")
        if r_cutoff is not None:
            self.r_cutoff = r_cutoff
        self.masks, self.rij, self.Rij = _calculate_masks_and_distances(
            structure, jnp.atleast_1d(self.r_cutoff)
        )

    def __hash__(self) -> int:
        """Enforce to use the parent class's hash method (JIT)."""
        return super().__hash__()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(r_cutoff={self.r_cutoff})"


register_jax_pytree_node(Neighbor)



from typing import Dict

from frozendict import frozendict
from jax import grad, jit
from pantea.atoms.structure import Inputs
from pantea.potentials._energy import AtomicPotentialInterface, _energy_fn
from pantea.types import Array, Element

_grad_energy_fn = jit(
    grad(_energy_fn, argnums=1),
    static_argnums=0,
)


def _compute_force(
    atomic_potential: Dict[Element, AtomicPotentialInterface],
    positions: Dict[Element, Array],
    params: Dict[Element, frozendict],
    inputs: Dict[Element, Inputs],
) -> Dict[Element, Array]:
    """Compute force components using the gradient of the total energy."""
    grad_energies: Dict[Element, Array] = _grad_energy_fn(
        atomic_potential,
        positions,
        params,
        inputs,
    )
    return {
        element: -1.0 * grad_energy
        for element, grad_energy in grad_energies.items()
    }



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, Optional, Protocol, Tuple

import jax.numpy as jnp
from frozendict import frozendict
from jax import random
from pantea.atoms.element import ElementMap
from pantea.atoms.structure import Structure
from pantea.datasets.dataset import DatasetInterface
from pantea.datasets.runner import RunnerDataset
from pantea.descriptors.acsf.acsf import ACSF
from pantea.descriptors.acsf.angular import G3, G9
from pantea.descriptors.acsf.cutoff import CutoffFunction
from pantea.descriptors.acsf.radial import G1, G2
from pantea.descriptors.scaler import Scaler
from pantea.logger import logger
from pantea.models.nn.initializer import UniformInitializer
from pantea.models.nn.network import NeuralNetworkModel
from pantea.potentials._energy import _compute_energy
from pantea.potentials._force import _compute_force
from pantea.potentials.atomic_potential import AtomicPotential
from pantea.potentials.nnp.gradient_descent import GradientDescentUpdater
from pantea.potentials.nnp.kalman_filter import KalmanFilterUpdater
from pantea.potentials.nnp.settings import (
    NeuralNetworkPotentialSettings as PotentialSettings,
)
from pantea.types import Array, Element
from tqdm import tqdm


class UpdaterInterface(Protocol):
    """Interface for potential weight updaters."""

    def fit(self, dataset: DatasetInterface) -> Dict:
        ...


@dataclass
class NeuralNetworkPotential:
    """
    High-dimensional neural network potential (HDNNP) - second generation.

    It contains all the required descriptors, scalers, and neural networks for each element,
    and an updater to fit the potential model using the reference structure data.

    Example
    -------
    Different methods to initialize neural network potential (NNP):

    .. code-block:: python
        :linenos:

        from pantea.potentials import NeuralNetworkPotential
        from pantea.potentials.nnp import NeuralNetworkPotentialSettings as Settings

        # Method #1 (potential file)
        nnp1 = NeuralNetworkPotential.from_file("input.nn")

        # Method #2 (json file)
        nnp3 = NeuralNetworkPotential.from_file('h2o.json')

        # Method #3 (dictionary of parameters)
        settings = Settings(**params_dict)
        nnp3 = NeuralNetworkPotential(settings)
    """

    settings: PotentialSettings = field(repr=False)
    output_dir: Path = field(default=Path("."), repr=False)
    atomic_potential: Dict[Element, AtomicPotential] = field(
        default_factory=dict, repr=True, init=False
    )
    model_params: Dict[Element, frozendict] = field(
        default_factory=dict, repr=False, init=False
    )
    updater: Optional[UpdaterInterface] = field(
        default=None, repr=False, init=False
    )

    @classmethod
    def from_file(cls, filename: Path) -> NeuralNetworkPotential:
        """Create an instance of the potential from input file (RuNNer format)."""
        logger.info("Creating potential from input file (RuNNer format)")
        potfile = Path(filename)
        suffix: str = potfile.suffix.lower()
        if ".json" == suffix:
            logger.info(f"Initializing from json file: {potfile.name}")
            settings = PotentialSettings.from_json(potfile)
        else:
            logger.info(
                f"Initializing from original RuNNer file format: {potfile.name}"
            )
            settings = PotentialSettings.from_file(potfile)

        return NeuralNetworkPotential(
            settings=settings,  # type: ignore
            output_dir=potfile.parent,
        )

    def __post_init__(self) -> None:
        """
        Initialize potential components such as atomic potential, model params,
        and updater from the potential settings.
        """
        logger.info(f"Initializing {self.__class__.__name__}")

        self.elements: Tuple[Element, ...] = tuple(
            element for element in self.settings.elements
        )
        self.num_elements: int = self.settings.number_of_elements

        logger.info(f"Number of elements: {self.num_elements}")
        for element in self.elements:
            logger.info(
                f"Element: {element} ({ElementMap.get_atomic_number(element)})"
            )

        if not self.atomic_potential:
            self._init_atomic_potential()
        if not self.model_params:
            self._init_model_params()
        if self.updater is None:
            self._init_updater()

    def _init_updater(self) -> None:
        """Initialize selected updater from the potential settings."""
        logger.info("Initializing updater")
        updater_type: str = self.settings.updater_type
        if updater_type == "kalman_filter":
            self.updater = KalmanFilterUpdater(potential=self)
        elif updater_type == "gradient_descent":
            self.updater = GradientDescentUpdater(potential=self)
        else:
            logger.error(
                f"Unknown updater type: {updater_type}",
                exception=TypeError,
            )
        logger.info(f"Updater type: {updater_type}")

    def _init_atomic_potential(self) -> None:
        """
        Initialize atomic potential for each element.
        This method can be override in case that different atomic potential is used.
        """
        logger.info("Initializing atomic potentials")
        descriptor: Dict[Element, ACSF] = self._init_descriptor()
        scaler: Dict[Element, Scaler] = self._init_scaler()
        model: Dict[Element, NeuralNetworkModel] = self._init_model()
        for element in self.settings.elements:
            self.atomic_potential[element] = AtomicPotential(
                descriptor=descriptor[element],
                scaler=scaler[element],
                model=model[element],
            )

    def _init_model_params(self) -> None:
        """
        Initialize neural network model parameters for each element
        (e.g. neural network kernel and bias values).

        This method be used to initialize model params of the potential with a different random seed.
        """
        logger.info("Initializing model params")
        random_keys = random.split(
            random.PRNGKey(self.settings.random_seed),
            self.settings.number_of_elements,
        )
        for i, element in enumerate(self.settings.elements):
            self.model_params[element] = self.atomic_potential[element].model.init(  # type: ignore
                random_keys[i],
                jnp.ones((1, self.atomic_potential[element].model_input_size)),
            )[
                "params"
            ]

    def _init_descriptor(self) -> Dict[Element, ACSF]:
        """Initialize descriptor for each element."""
        logger.info("Initializing descriptors")
        descriptor: Dict[Element, ACSF] = dict()
        settings = self.settings
        # Instantiate ACSF for each element
        for element in settings.elements:
            descriptor[element] = ACSF(element)
        # Add symmetry functions
        logger.debug("Registering symmetry functions (radial and angular)")

        for args in settings.symfunction_short:
            if args.acsf_type == 1:
                descriptor[args.central_element].add(
                    symmetry_function=G1(
                        CutoffFunction.from_cutoff_type(
                            r_cutoff=args.r_cutoff,
                            cutoff_type=settings.cutoff_type,
                        )
                    ),
                    neighbor_element_j=args.neighbor_element_j,
                )
            elif args.acsf_type == 2:
                descriptor[args.central_element].add(
                    symmetry_function=G2(
                        CutoffFunction.from_cutoff_type(
                            r_cutoff=args.r_cutoff,
                            cutoff_type=settings.cutoff_type,
                        ),
                        eta=args.eta,
                        r_shift=args.r_shift,
                    ),
                    neighbor_element_j=args.neighbor_element_j,
                )
            elif args.acsf_type == 3:
                descriptor[args.central_element].add(
                    symmetry_function=G3(
                        CutoffFunction.from_cutoff_type(
                            r_cutoff=args.r_cutoff,
                            cutoff_type=settings.cutoff_type,
                        ),
                        eta=args.eta,
                        zeta=args.zeta,  # type: ignore
                        lambda0=args.lambda0,  # type: ignore
                        r_shift=args.r_cutoff,
                    ),
                    neighbor_element_j=args.neighbor_element_j,
                    neighbor_element_k=args.neighbor_element_k,  # type: ignore
                )
            elif args.acsf_type == 9:
                descriptor[args.central_element].add(
                    symmetry_function=G9(
                        CutoffFunction.from_cutoff_type(
                            r_cutoff=args.r_cutoff,
                            cutoff_type=settings.cutoff_type,
                        ),
                        eta=args.eta,
                        zeta=args.zeta,  # type: ignore
                        lambda0=args.lambda0,  # type: ignore
                        r_shift=args.r_shift,
                    ),
                    neighbor_element_j=args.neighbor_element_j,
                    neighbor_element_k=args.neighbor_element_k,  # type: ignore
                )
        return descriptor

    def _init_scaler(self) -> Dict[Element, Scaler]:
        """Initialize descriptor scaler for each element."""
        logger.info("Initializing descriptor scalers")
        scaler: Dict[Element, Scaler] = dict()
        settings = self.settings
        # Prepare scaler input argument if exist in settings
        scaler_kwargs = {
            first: settings[second]
            for first, second in {
                "scale_type": "scale_type",
                "scale_min": "scale_min_short",
                "scale_max": "scale_max_short",
            }.items()
            if second in self.settings.keywords()
        }
        logger.debug(f"Scaler kwargs={scaler_kwargs}")
        # Assign an ACSF scaler to each element
        for element in settings.elements:
            scaler[element] = Scaler(**scaler_kwargs)
        return scaler

    def _init_model(self) -> Dict[Element, NeuralNetworkModel]:
        """Initialize neural network model for each element."""
        logger.info("Initializing neural network models")
        model: Dict[Element, NeuralNetworkModel] = dict()
        settings = self.settings

        for element in settings.elements:
            logger.debug(f"Element: {element}")

            hidden_layers = zip(
                settings.global_nodes_short,
                settings.global_activation_short[:-1],
            )
            output_layer: Tuple[int, str] = (
                1,
                settings.global_activation_short[-1],
            )
            kernel_initializer: UniformInitializer = UniformInitializer(
                weights_range=(
                    settings.weights_min,
                    settings.weights_max,
                )
            )
            model[element] = NeuralNetworkModel(
                hidden_layers=tuple([(n, t) for n, t in hidden_layers]),
                output_layer=output_layer,
                kernel_initializer=kernel_initializer,
            )
        return model

    # ------------------------------------------------------------------------

    def __call__(self, structure: Structure) -> Array:
        """
        Compute the total energy.

        :param structure: Structure
        :return: total energy
        """

        return _compute_energy(
            frozendict(self.atomic_potential),  # must be hashable
            structure.get_per_element_positions(),
            self.model_params,
            structure.get_per_element_inputs(),
        )

    def compute_forces(self, structure: Structure) -> Array:
        """
        Compute force components.

        :param structure: input structure
        :return: predicted force components for all atoms
        """
        force_dict: Dict[Element, Array] = _compute_force(
            frozendict(self.atomic_potential),  # must be hashable
            structure.get_per_element_positions(),
            self.model_params,
            structure.get_per_element_inputs(),
        )
        force: Array = jnp.empty_like(structure.forces)
        for element in structure.get_unique_elements():
            atom_index: Array = structure.select(element)
            force = force.at[atom_index].set(force_dict[element])

        return force

    # ------------------------------------------------------------------------

    def fit_scaler(self, dataset: RunnerDataset) -> None:
        """
        Fit scaler parameters for each element using the input structure data.
        No gradient history is required here.
        """
        print("Fitting descriptor scaler...")
        try:
            dataset_size: int = len(dataset)
            for index in tqdm(range(dataset_size)):
                structure: Structure = dataset[index]
                elements: Tuple[Element, ...] = structure.get_unique_elements()
                for element in elements:
                    x: Array = self.atomic_potential[element].descriptor(
                        structure
                    )
                    self.atomic_potential[element].scaler.fit(x)
        except KeyboardInterrupt:
            print("Keyboard Interrupt")
        else:
            print("Done.")

    def fit_model(self, dataset: RunnerDataset) -> Dict:
        """
        Fit energy model for all elements using the input structure loader.
        """
        # kwargs["validation_split"] = kwargs.get(
        #     "validation_split", self.settings.test_fraction
        # )
        # kwargs["epochs"] = kwargs.get("epochs", self.settings.epochs)

        history = defaultdict(list)
        print("Training potential...")
        try:
            history = self.updater.fit(dataset)  # type: ignore
        except KeyboardInterrupt:
            print("Keyboard Interrupt")
        else:
            print("Done.")

        return history

    def fit(self) -> None:
        """
        This method provides a user-friendly interface to fit both descriptor and model in one step.
        """
        ...

    # ------------------------------------------------------------------------

    def save_scaler(self) -> None:
        """
        This method saves scaler parameters for each element into separate files.
        """
        # Save scaler parameters for each element separately
        for element in self.settings.elements:
            atomic_number: int = ElementMap.get_atomic_number(element)
            scaler_file = Path(
                self.output_dir,
                self.settings.scaler_save_naming_format.format(atomic_number),
            )
            logger.info(
                f"Saving scaler parameters for element ({element}): {scaler_file.name}"
            )
            self.atomic_potential[element].scaler.save(scaler_file)

    def save_model(self) -> None:
        """
        Save model weights separately for all elements.
        """
        for element in self.elements:
            atomic_number = ElementMap.get_atomic_number(element)
            model_file = Path(
                self.output_dir,
                self.settings.model_save_naming_format.format(atomic_number),
            )
            logger.info(
                f"Saving model weights for element ({element}): {model_file.name}"
            )
            self.atomic_potential[element].model.save(
                model_file, self.model_params[element]
            )

    def save(self) -> None:
        """Save scaler and model into corresponding files for each element."""
        self.save_scaler()
        self.save_model()

    def load_scaler(self) -> None:
        """
        This method loads scaler parameters of each element from separate files.
        This save computational time as the would be no need to fit the scalers each time.
        """
        # Load scaler parameters for each element separately
        for element in self.elements:
            atomic_number: int = ElementMap.get_atomic_number(element)
            scaler_file = Path(
                self.output_dir,
                self.settings.scaler_save_naming_format.format(atomic_number),
            )
            logger.info(
                f"Loading scaler parameters for element ({element}): {scaler_file.name}"
            )
            self.atomic_potential[element].scaler.load(scaler_file)

    def load_model(self) -> None:
        """
        Load model weights separately for all elements.
        """
        for element in self.elements:
            atomic_number: int = ElementMap.get_atomic_number(element)
            model_file = Path(
                self.output_dir,
                self.settings.model_save_naming_format.format(atomic_number),
            )
            logger.info(
                f"Loading model weights for element ({element}): {model_file.name}"
            )
            self.model_params[element] = self.atomic_potential[
                element
            ].model.load(model_file)

    def load(self) -> None:
        """Load element scaler and model from their corresponding files."""
        self.load_scaler()
        self.load_model()

    def set_extrapolation_warnings(
        self, threshold: Optional[int] = None
    ) -> None:
        """
        shows warning whenever a descriptor value is out of bounds defined by
        minimum/maximum values in the scaler.

        set_extrapolation_warnings(None) will disable it.

        :param threshold: maximum number of warnings
        :type threshold: int
        """
        logger.info(f"Setting extrapolation warning: {threshold}")
        for pot in self.atomic_potential.values():
            pot.scaler.set_max_number_of_warnings(threshold)

    @property
    def extrapolation_warnings(self) -> Dict[Element, int]:
        return {
            element: pot.scaler.number_of_warnings
            for element, pot in self.atomic_potential.items()
        }

    @property
    def r_cutoff(self) -> float:
        """Return the maximum cutoff radius found between all descriptors."""
        return max(
            [pot.descriptor.r_cutoff for pot in self.atomic_potential.values()]
        )

    @property
    def descriptor(self) -> Dict:
        """Return descriptor for each element."""
        return {
            elem: pot.descriptor for elem, pot in self.atomic_potential.items()
        }

    @property
    def scaler(self) -> Dict:
        """Return scaler for each element."""
        return {
            elem: pot.scaler for elem, pot in self.atomic_potential.items()
        }

    @property
    def model(self) -> Dict:
        """Return model for each element."""
        return {elem: pot.model for elem, pot in self.atomic_potential.items()}


NNP = NeuralNetworkPotential



from __future__ import annotations

import math
from dataclasses import dataclass
from functools import partial, update_wrapper
from typing import Any, Callable, Mapping

import jax
import jax.numpy as jnp
from pantea.pytree import BaseJaxPytreeDataClass, register_jax_pytree_node
from pantea.types import Array

_TANH_PRE: float = ((math.e + 1 / math.e) / (math.e - 1 / math.e)) ** 3


def _hard(r: Array) -> Array:
    return jnp.ones_like(r)


def _tanhu(r: Array, r_cutoff: float) -> Array:
    return jnp.tanh(1.0 - r / r_cutoff) ** 3


def _tanh(r: Array, r_cutoff: float) -> Array:
    return _TANH_PRE * jnp.tanh(1.0 - r / r_cutoff) ** 3


def _cos(r: Array, r_cutoff: float) -> Array:
    return 0.5 * (jnp.cos(jnp.pi * r / r_cutoff) + 1.0)


def _exp(r: Array, r_cutoff: float) -> Array:
    return jnp.exp(1.0 - 1.0 / (1.0 - (r / r_cutoff) ** 2))


def _poly1(r: Array) -> Array:
    return (2.0 * r - 3.0) * r**2 + 1.0


def _poly2(r: Array) -> Array:
    return ((15.0 - 6.0 * r) * r - 10) * r**3 + 1.0


def _wrapped_partial(function: Callable, **kwargs: Any) -> Callable:
    partial_function = partial(function, **kwargs)
    update_wrapper(partial_function, function)
    return partial_function


@dataclass(frozen=True)
class CutoffFunction(BaseJaxPytreeDataClass):
    """Cutoff function for ACSF descriptor.

    Cutoff functions are utilized in the calculation of Atom-centered Symmetry
    Function (ACSF) descriptors. These functions serve to limit the influence
    of atoms located beyond a specified distance from the central atom.

    The ACSF descriptors employ cutoff functions to determine the range within which
    neighboring atoms contribute to the descriptor calculation. In fact,
    cutoff function assigns a weight to each neighbor atom based on its
    distance from the central atom. Typically, a smooth cutoff function is
    employed to smoothly taper off the contribution of
    atoms as they move away from the central atom.

    The choice of cutoff function can vary depending on the specific application.
    Examples of commonly used cutoff functions include the hyperbolic tangent (tanh)
    cutoff, exponential, or exponential.


    See `cutoff function`_ and `cutoff type`_ for more details.

    .. _`cutoff function`: https://compphysvienna.github.io/n2p2/api/cutoff_functions.html?highlight=cutoff#
    .. _`cutoff type`: https://compphysvienna.github.io/n2p2/topics/keywords.html?highlight=cutoff_type
    """

    r_cutoff: float
    cutoff_function: Callable

    @classmethod
    def from_cutoff_type(
        cls,
        r_cutoff: float,
        cutoff_type: str = "tanh",
    ) -> CutoffFunction:
        """Create cutoff function from the input cutoff type such as "tanh", "cos", and "tanh"."""
        _cutoff_function_map: Mapping[str, Callable] = {
            "hard": _hard,
            "tanhu": _wrapped_partial(_tanhu, r_cutoff=r_cutoff),
            "tanh": _wrapped_partial(_tanh, r_cutoff=r_cutoff),
            "cos": _wrapped_partial(_cos, r_cutoff=r_cutoff),
            "exp": _wrapped_partial(_exp, r_cutoff=r_cutoff),
            "poly1": _poly1,
            "poly2": _poly2,
        }
        return cls(r_cutoff, _cutoff_function_map[cutoff_type])

    def __post_init__(self) -> None:
        self._assert_jit_dynamic_attributes()
        self._assert_jit_static_attributes(
            expected=("r_cutoff", "cutoff_function")
        )

    @jax.jit
    def __call__(self, r: Array) -> Array:
        return jnp.where(
            r < self.r_cutoff,
            self.cutoff_function(r),
            jnp.zeros_like(r),
        )

    def __hash__(self) -> int:
        """Override the hash function from the base jax pytree data class."""
        return super().__hash__()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(r_cutoff={self.r_cutoff})"


register_jax_pytree_node(CutoffFunction)



from pantea.descriptors.acsf.acsf import ACSF
from pantea.descriptors.acsf.angular import G3, G9, AngularSymmetryFunction
from pantea.descriptors.acsf.cutoff import CutoffFunction
from pantea.descriptors.acsf.radial import G1, G2, RadialSymmetryFunction

__all__ = [
    "ACSF",
    "CutoffFunction",
    "RadialSymmetryFunction",
    "AngularSymmetryFunction",
    "G1",
    "G2",
    "G3",
    "G9",
]



from __future__ import annotations

from collections import defaultdict
from pathlib import Path
from typing import Dict, Iterator, List, Optional, TextIO

from pantea.atoms.structure import Structure
from pantea.datasets.dataset import DatasetInterface
from pantea.logger import logger
from pantea.types import Dtype, default_dtype
from pantea.utils.tokenize import tokenize


class RunnerDataset(DatasetInterface):
    """

    The dataset used for the input data format of `RuNNer`_ consists of atomic attributes
    and simulation box information. Within each snapshot, there are two types of
    properties: `per-atom` properties and `collective` properties.

    The per-atom properties encompass various attributes like the element name,
    positions, energy, charge, force components, and more.

    On the other hand, the collective properties include attributes
    such as lattice parameters, total energy, and total charge.

    .. _RuNNer: https://www.uni-goettingen.de/de/560580.html
    """

    def __init__(
        self,
        filename: Path,
        persist: bool = False,
        dtype: Optional[Dtype] = None,
    ) -> None:
        """
        Create a `RuNNer`_ structure dataset by initializing it from an input file.

        :param filename: input file name
        :type filename: Path
        :param persist: Persist any loaded structure data in the memory, defaults to False
        :type persist: bool, optional
        :param dtype: floating point precision for the structure data, defaults to None
        :type dtype: Optional[Dtype], optional

        .. _RuNNer: https://www.uni-goettingen.de/de/560580.html
        """
        self.filename: Path = Path(filename)
        self.persist: bool = persist
        self._cache: Dict[int, Structure] = dict()
        self.dtype = dtype if dtype is not None else default_dtype.FLOATX

    def __len__(self) -> int:
        """Return number of available structures."""
        num_structures: int = 0
        with open(str(self.filename), "r") as file:
            while self._ignore_next(file):
                num_structures += 1
        return num_structures

    def __getitem__(self, index: int) -> Structure:
        """
        Return i-th structure.

        This is a lazy call which means that only required section
        of data is loaded into the memory.
        """
        return self._read_from_cache(index)

    def read_sequential(self) -> Iterator[Structure]:
        """
        Read structures sequentially.

        It must be noted that reading data in a consecutive manner is
        faster compared to random indexing read.

        :return: Structure
        :rtype: Iterator[Structure]
        """
        logger.debug("Read structures sequentially")
        index: int = 0
        with open(str(self.filename), "r") as file:
            while True:
                data = self._read_next(file)
                if not data:
                    break
                structure = self._to_structure(data)
                if self.persist:
                    if index in self._cache:
                        yield self._cache[index]
                    else:
                        self._cache[index] = structure
                yield structure
                index += 1

    def preload(self) -> None:
        """
        Preload (cache) all structures into memory.

        This ensures that any structure can be rapidly loaded from memory in subsequent operations.
        """
        self.persist = True
        for _ in self.read_sequential():
            pass

    @classmethod
    def _read_next(cls, file: TextIO) -> Dict[str, List]:
        """Read next structure data between `begin` and `end` keywords."""
        data = defaultdict(list)
        read_block: bool = False
        while True:
            line = file.readline()
            if not line:
                break
            keyword, _ = tokenize(line)
            if keyword == "begin":
                read_block = True
                break
        while read_block:
            line = file.readline()
            if not line:
                break
            keyword, tokens = tokenize(line)
            if keyword == "atom":
                data["positions"].append([float(t) for t in tokens[:3]])
                data["elements"].append(tokens[3])
                data["charges"].append(float(tokens[4]))
                data["energies"].append(float(tokens[5]))
                data["forces"].append([float(t) for t in tokens[6:9]])
            elif keyword == "lattice":
                data["lattice"].append([float(t) for t in tokens[:3]])
            elif keyword == "energy":
                data["total_energy"].append(float(tokens[0]))
            elif keyword == "charge":
                data["total_charge"].append(float(tokens[0]))
            elif keyword == "comment":
                data["comment"].append(" ".join(line.split()[1:]))
            elif keyword == "end":
                read_block = False
        return data

    @classmethod
    def _ignore_next(cls, file: TextIO) -> bool:
        while True:
            line = file.readline()
            if not line:
                return False
            keyword, _ = tokenize(line)
            if keyword == "end":
                break
        return True

    def _read(self, index: int) -> Structure:
        logger.debug(f"load structure({index=})")
        with open(str(self.filename), "r") as file:
            for _ in range(index):
                self._ignore_next(file)
            data = self._read_next(file)
            if not data:
                raise IndexError(
                    f"The given index {index} is out of bound (len={len(self)})"
                )
        return self._to_structure(data)

    def _to_structure(self, data: Dict[str, List]) -> Structure:
        return Structure.from_dict(data, dtype=self.dtype)

    def _read_from_cache(self, index: int) -> Structure:
        """Read the desired structure from cache, if possible."""
        if not self.persist:
            return self._read(index)
        if index not in self._cache:
            structure = self._read(index)
            self._cache[index] = structure
            return structure
        else:
            return self._cache[index]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}"
            f"(filename='{str(self.filename)}'"
            f", persist={self.persist}, dtype={self.dtype.dtype})"
        )



import os
from typing import Tuple

os.environ["JAX_ENABLE_X64"] = "1"
os.environ["JAX_PLATFORM_NAME"] = "cpu"

import jax
import jax.numpy as jnp
import pytest
from frozendict import frozendict
from pantea.models.nn.initializer import UniformInitializer
from pantea.models.nn.network import NeuralNetworkModel
from pantea.types import default_dtype


class TestNeuralNetworkModel:
    nn_1: NeuralNetworkModel = NeuralNetworkModel(hidden_layers=((2, "tanh"),))
    nn_2: NeuralNetworkModel = NeuralNetworkModel(
        hidden_layers=((4, "tanh"), (8, "identity")),
        param_dtype=jnp.float64,  # type: ignore
    )
    nn_3: NeuralNetworkModel = NeuralNetworkModel(
        hidden_layers=((8, "tanh"), (16, "tanh")),
        kernel_initializer=UniformInitializer(weights_range=(-1.0, 1.0)),
    )
    nn_4: NeuralNetworkModel = NeuralNetworkModel(
        hidden_layers=((8, "tanh"), (16, "tanh")),
        kernel_initializer=UniformInitializer(weights_range=(0.0, 1.0)),
    )

    @pytest.mark.parametrize(
        "network, expected",
        [
            (
                nn_1,
                (
                    ((2, "tanh"),),
                    (1, "identity"),
                    default_dtype.FLOATX,
                ),
            ),
            (
                nn_2,
                (
                    ((4, "tanh"), (8, "identity")),
                    (1, "identity"),
                    jnp.float64,
                ),
            ),
            (
                nn_3,
                (
                    ((8, "tanh"), (16, "tanh")),
                    (1, "identity"),
                    default_dtype.FLOATX,
                ),
            ),
        ],
    )
    def test_attributes(
        self,
        network: NeuralNetworkModel,
        expected: Tuple,
    ) -> None:
        assert network.hidden_layers == expected[0]
        assert network.output_layer == expected[1]
        assert network.param_dtype, expected[2]

    @pytest.mark.parametrize(
        "network, expected",
        [
            (
                nn_3,
                (
                    UniformInitializer,
                    (-1.0, 1.0),
                ),
            ),
            (
                nn_4,
                (
                    UniformInitializer,
                    (0.0, 1.0),
                ),
            ),
        ],
    )
    def test_kernel_initializers(
        self,
        network: NeuralNetworkModel,
        expected: Tuple,
    ) -> None:
        assert isinstance(network.kernel_initializer, expected[0])
        assert network.kernel_initializer.weights_range == expected[1]

    @pytest.mark.parametrize(
        "network, inputs, expected",
        [
            (
                nn_3,
                (4, 5),
                (
                    frozendict(
                        {
                            "params": {
                                "layers_0": {
                                    "bias": (8,),
                                    "kernel": (5, 8),
                                },
                                "layers_2": {
                                    "bias": (16,),
                                    "kernel": (8, 16),
                                },
                                "layers_4": {
                                    "bias": (1,),
                                    "kernel": (16, 1),
                                },
                            },
                        }
                    ),
                ),
            ),
        ],
    )
    def test_network_creation(
        self,
        network: NeuralNetworkModel,
        inputs: Tuple,
        expected: Tuple,
    ) -> None:
        rng = jax.random.PRNGKey(2023)  # PRNG Key
        x = jnp.ones(shape=(inputs[0], inputs[1]))  # Dummy Input
        params = network.init(rng, x)  # type: ignore
        y = network.apply(params, x)
        assert y.shape[0] == inputs[0]  # type: ignore
        assert y.shape[1] == network.output_layer[0]  # type: ignore
        assert jax.tree_map(lambda x: x.shape, params) == expected[0]
